# orm/interfaces.py
# Copyright (C) 2005-2023 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: https://www.opensource.org/licenses/mit-license.php

"""

Contains various base classes used throughout the ORM.

Defines some key base classes prominent within the internals.

This module and the classes within are mostly private, though some attributes
are exposed when inspecting mappings.

"""

from __future__ import annotations

import collections
import dataclasses
import typing
from typing import Any
from typing import Callable
from typing import cast
from typing import ClassVar
from typing import Dict
from typing import Generic
from typing import Iterator
from typing import List
from typing import NamedTuple
from typing import NoReturn
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import Type
from typing import TYPE_CHECKING
from typing import TypeVar
from typing import Union

from . import exc as orm_exc
from . import path_registry
from .base import _MappedAttribute as _MappedAttribute
from .base import EXT_CONTINUE as EXT_CONTINUE  # noqa: F401
from .base import EXT_SKIP as EXT_SKIP  # noqa: F401
from .base import EXT_STOP as EXT_STOP  # noqa: F401
from .base import InspectionAttr as InspectionAttr  # noqa: F401
from .base import InspectionAttrInfo as InspectionAttrInfo
from .base import MANYTOMANY as MANYTOMANY  # noqa: F401
from .base import MANYTOONE as MANYTOONE  # noqa: F401
from .base import NO_KEY as NO_KEY  # noqa: F401
from .base import NO_VALUE as NO_VALUE  # noqa: F401
from .base import NotExtension as NotExtension  # noqa: F401
from .base import ONETOMANY as ONETOMANY  # noqa: F401
from .base import RelationshipDirection as RelationshipDirection  # noqa: F401
from .base import SQLORMOperations
from .. import ColumnElement
from .. import exc as sa_exc
from .. import inspection
from .. import util
from ..sql import operators
from ..sql import roles
from ..sql import visitors
from ..sql.base import _NoArg
from ..sql.base import ExecutableOption
from ..sql.cache_key import HasCacheKey
from ..sql.operators import ColumnOperators
from ..sql.schema import Column
from ..sql.type_api import TypeEngine
from ..util import warn_deprecated
from ..util.typing import RODescriptorReference
from ..util.typing import TypedDict

if typing.TYPE_CHECKING:
    from ._typing import _EntityType
    from ._typing import _IdentityKeyType
    from ._typing import _InstanceDict
    from ._typing import _InternalEntityType
    from ._typing import _ORMAdapterProto
    from .attributes import InstrumentedAttribute
    from .base import Mapped
    from .context import _MapperEntity
    from .context import ORMCompileState
    from .context import QueryContext
    from .decl_api import RegistryType
    from .decl_base import _ClassScanMapperConfig
    from .loading import _PopulatorDict
    from .mapper import Mapper
    from .path_registry import AbstractEntityRegistry
    from .query import Query
    from .session import Session
    from .state import InstanceState
    from .strategy_options import _LoadElement
    from .util import AliasedInsp
    from .util import ORMAdapter
    from ..engine.result import Result
    from ..sql._typing import _ColumnExpressionArgument
    from ..sql._typing import _ColumnsClauseArgument
    from ..sql._typing import _DMLColumnArgument
    from ..sql._typing import _InfoType
    from ..sql.operators import OperatorType
    from ..sql.visitors import _TraverseInternalsType
    from ..util.typing import _AnnotationScanType

_StrategyKey = Tuple[Any, ...]

_T = TypeVar("_T", bound=Any)
_T_co = TypeVar("_T_co", bound=Any, covariant=True)

_TLS = TypeVar("_TLS", bound="Type[LoaderStrategy]")


class ORMStatementRole(roles.StatementRole):
    __slots__ = ()
    _role_name = (
        "Executable SQL or text() construct, including ORM " "aware objects"
    )


class ORMColumnsClauseRole(
    roles.ColumnsClauseRole, roles.TypedColumnsClauseRole[_T]
):
    __slots__ = ()
    _role_name = "ORM mapped entity, aliased entity, or Column expression"


class ORMEntityColumnsClauseRole(ORMColumnsClauseRole[_T]):
    __slots__ = ()
    _role_name = "ORM mapped or aliased entity"


class ORMFromClauseRole(roles.StrictFromClauseRole):
    __slots__ = ()
    _role_name = "ORM mapped entity, aliased entity, or FROM expression"


class ORMColumnDescription(TypedDict):
    name: str
    # TODO: add python_type and sql_type here; combining them
    # into "type" is a bad idea
    type: Union[Type[Any], TypeEngine[Any]]
    aliased: bool
    expr: _ColumnsClauseArgument[Any]
    entity: Optional[_ColumnsClauseArgument[Any]]


class _IntrospectsAnnotations:
    __slots__ = ()

    def found_in_pep593_annotated(self) -> Any:
        """return a copy of this object to use in declarative when the
        object is found inside of an Annotated object."""

        raise NotImplementedError(
            f"Use of the {self.__class__} construct inside of an "
            f"Annotated object is not yet supported."
        )

    def declarative_scan(
        self,
        decl_scan: _ClassScanMapperConfig,
        registry: RegistryType,
        cls: Type[Any],
        originating_module: Optional[str],
        key: str,
        mapped_container: Optional[Type[Mapped[Any]]],
        annotation: Optional[_AnnotationScanType],
        extracted_mapped_annotation: Optional[_AnnotationScanType],
        is_dataclass_field: bool,
    ) -> None:
        """Perform class-specific initializaton at early declarative scanning
        time.

        .. versionadded:: 2.0

        """

    def _raise_for_required(self, key: str, cls: Type[Any]) -> NoReturn:
        raise sa_exc.ArgumentError(
            f"Python typing annotation is required for attribute "
            f'"{cls.__name__}.{key}" when primary argument(s) for '
            f'"{self.__class__.__name__}" construct are None or not present'
        )


class _AttributeOptions(NamedTuple):
    """define Python-local attribute behavior options common to all
    :class:`.MapperProperty` objects.

    Currently this includes dataclass-generation arguments.

    .. versionadded:: 2.0

    """

    dataclasses_init: Union[_NoArg, bool]
    dataclasses_repr: Union[_NoArg, bool]
    dataclasses_default: Union[_NoArg, Any]
    dataclasses_default_factory: Union[_NoArg, Callable[[], Any]]
    dataclasses_compare: Union[_NoArg, bool]
    dataclasses_kw_only: Union[_NoArg, bool]

    def _as_dataclass_field(self, key: str) -> Any:
        """Return a ``dataclasses.Field`` object given these arguments."""

        kw: Dict[str, Any] = {}
        if self.dataclasses_default_factory is not _NoArg.NO_ARG:
            kw["default_factory"] = self.dataclasses_default_factory
        if self.dataclasses_default is not _NoArg.NO_ARG:
            kw["default"] = self.dataclasses_default
        if self.dataclasses_init is not _NoArg.NO_ARG:
            kw["init"] = self.dataclasses_init
        if self.dataclasses_repr is not _NoArg.NO_ARG:
            kw["repr"] = self.dataclasses_repr
        if self.dataclasses_compare is not _NoArg.NO_ARG:
            kw["compare"] = self.dataclasses_compare
        if self.dataclasses_kw_only is not _NoArg.NO_ARG:
            kw["kw_only"] = self.dataclasses_kw_only

        if "default" in kw and callable(kw["default"]):
            # callable defaults are ambiguous. deprecate them in favour of
            # insert_default or default_factory. #9936
            warn_deprecated(
                f"Callable object passed to the ``default`` parameter for "
                f"attribute {key!r} in a ORM-mapped Dataclasses context is "
                "ambiguous, "
                "and this use will raise an error in a future release.  "
                "If this callable is intended to produce Core level INSERT "
                "default values for an underlying ``Column``, use "
                "the ``mapped_column.insert_default`` parameter instead.  "
                "To establish this callable as providing a default value "
                "for instances of the dataclass itself, use the "
                "``default_factory`` dataclasses parameter.",
                "2.0",
            )

        if (
            "init" in kw
            and not kw["init"]
            and "default" in kw
            and not callable(kw["default"])  # ignore callable defaults. #9936
            and "default_factory" not in kw  # illegal but let dc.field raise
        ):
            # fix for #9879
            default = kw.pop("default")
            kw["default_factory"] = lambda: default

        return dataclasses.field(**kw)

    @classmethod
    def _get_arguments_for_make_dataclass(
        cls,
        key: str,
        annotation: _AnnotationScanType,
        mapped_container: Optional[Any],
        elem: _T,
    ) -> Union[
        Tuple[str, _AnnotationScanType],
        Tuple[str, _AnnotationScanType, dataclasses.Field[Any]],
    ]:
        """given attribute key, annotation, and value from a class, return
        the argument tuple we would pass to dataclasses.make_dataclass()
        for this attribute.

        """
        if isinstance(elem, _DCAttributeOptions):
            dc_field = elem._attribute_options._as_dataclass_field(key)

            return (key, annotation, dc_field)
        elif elem is not _NoArg.NO_ARG:
            # why is typing not erroring on this?
            return (key, annotation, elem)
        elif mapped_container is not None:
            # it's Mapped[], but there's no "element", which means declarative
            # did not actually do anything for this field.  this shouldn't
            # happen.
            # previously, this would occur because _scan_attributes would
            # skip a field that's on an already mapped superclass, but it
            # would still include it in the annotations, leading
            # to issue #8718

            assert False, "Mapped[] received without a mapping declaration"

        else:
            # plain dataclass field, not mapped.  Is only possible
            # if __allow_unmapped__ is set up.  I can see this mode causing
            # problems...
            return (key, annotation)


_DEFAULT_ATTRIBUTE_OPTIONS = _AttributeOptions(
    _NoArg.NO_ARG,
    _NoArg.NO_ARG,
    _NoArg.NO_ARG,
    _NoArg.NO_ARG,
    _NoArg.NO_ARG,
    _NoArg.NO_ARG,
)

_DEFAULT_READONLY_ATTRIBUTE_OPTIONS = _AttributeOptions(
    False,
    _NoArg.NO_ARG,
    _NoArg.NO_ARG,
    _NoArg.NO_ARG,
    _NoArg.NO_ARG,
    _NoArg.NO_ARG,
)


class _DCAttributeOptions:
    """mixin for descriptors or configurational objects that include dataclass
    field options.

    This includes :class:`.MapperProperty`, :class:`._MapsColumn` within
    the ORM, but also includes :class:`.AssociationProxy` within ext.
    Can in theory be used for other descriptors that serve a similar role
    as association proxy.   (*maybe* hybrids, not sure yet.)

    """

    __slots__ = ()

    _attribute_options: _AttributeOptions
    """behavioral options for ORM-enabled Python attributes

    .. versionadded:: 2.0

    """

    _has_dataclass_arguments: bool


class _MapsColumns(_DCAttributeOptions, _MappedAttribute[_T]):
    """interface for declarative-capable construct that delivers one or more
    Column objects to the declarative process to be part of a Table.
    """

    __slots__ = ()

    @property
    def mapper_property_to_assign(self) -> Optional[MapperProperty[_T]]:
        """return a MapperProperty to be assigned to the declarative mapping"""
        raise NotImplementedError()

    @property
    def columns_to_assign(self) -> List[Tuple[Column[_T], int]]:
        """A list of Column objects that should be declaratively added to the
        new Table object.

        """
        raise NotImplementedError()


# NOTE: MapperProperty needs to extend _MappedAttribute so that declarative
# typing works, i.e. "Mapped[A] = relationship()".   This introduces an
# inconvenience which is that all the MapperProperty objects are treated
# as descriptors by typing tools, which are misled by this as assignment /
# access to a descriptor attribute wants to move through __get__.
# Therefore, references to MapperProperty as an instance variable, such
# as in PropComparator, may have some special typing workarounds such as the
# use of sqlalchemy.util.typing.DescriptorReference to avoid mis-interpretation
# by typing tools
@inspection._self_inspects
class MapperProperty(
    HasCacheKey,
    _DCAttributeOptions,
    _MappedAttribute[_T],
    InspectionAttrInfo,
    util.MemoizedSlots,
):
    """Represent a particular class attribute mapped by :class:`_orm.Mapper`.

    The most common occurrences of :class:`.MapperProperty` are the
    mapped :class:`_schema.Column`, which is represented in a mapping as
    an instance of :class:`.ColumnProperty`,
    and a reference to another class produced by :func:`_orm.relationship`,
    represented in the mapping as an instance of
    :class:`.Relationship`.

    """

    __slots__ = (
        "_configure_started",
        "_configure_finished",
        "_attribute_options",
        "_has_dataclass_arguments",
        "parent",
        "key",
        "info",
        "doc",
    )

    _cache_key_traversal: _TraverseInternalsType = [
        ("parent", visitors.ExtendedInternalTraversal.dp_has_cache_key),
        ("key", visitors.ExtendedInternalTraversal.dp_string),
    ]

    if not TYPE_CHECKING:
        cascade = None

    is_property = True
    """Part of the InspectionAttr interface; states this object is a
    mapper property.

    """

    comparator: PropComparator[_T]
    """The :class:`_orm.PropComparator` instance that implements SQL
    expression construction on behalf of this mapped attribute."""

    key: str
    """name of class attribute"""

    parent: Mapper[Any]
    """the :class:`.Mapper` managing this property."""

    _is_relationship = False

    _links_to_entity: bool
    """True if this MapperProperty refers to a mapped entity.

    Should only be True for Relationship, False for all others.

    """

    doc: Optional[str]
    """optional documentation string"""

    info: _InfoType
    """Info dictionary associated with the object, allowing user-defined
    data to be associated with this :class:`.InspectionAttr`.

    The dictionary is generated when first accessed.  Alternatively,
    it can be specified as a constructor argument to the
    :func:`.column_property`, :func:`_orm.relationship`, or :func:`.composite`
    functions.

    .. seealso::

        :attr:`.QueryableAttribute.info`

        :attr:`.SchemaItem.info`

    """

    def _memoized_attr_info(self) -> _InfoType:
        """Info dictionary associated with the object, allowing user-defined
        data to be associated with this :class:`.InspectionAttr`.

        The dictionary is generated when first accessed.  Alternatively,
        it can be specified as a constructor argument to the
        :func:`.column_property`, :func:`_orm.relationship`, or
        :func:`.composite`
        functions.

        .. seealso::

            :attr:`.QueryableAttribute.info`

            :attr:`.SchemaItem.info`

        """
        return {}

    def setup(
        self,
        context: ORMCompileState,
        query_entity: _MapperEntity,
        path: AbstractEntityRegistry,
        adapter: Optional[ORMAdapter],
        **kwargs: Any,
    ) -> None:
        """Called by Query for the purposes of constructing a SQL statement.

        Each MapperProperty associated with the target mapper processes the
        statement referenced by the query context, adding columns and/or
        criterion as appropriate.

        """

    def create_row_processor(
        self,
        context: ORMCompileState,
        query_entity: _MapperEntity,
        path: AbstractEntityRegistry,
        mapper: Mapper[Any],
        result: Result[Any],
        adapter: Optional[ORMAdapter],
        populators: _PopulatorDict,
    ) -> None:
        """Produce row processing functions and append to the given
        set of populators lists.

        """

    def cascade_iterator(
        self,
        type_: str,
        state: InstanceState[Any],
        dict_: _InstanceDict,
        visited_states: Set[InstanceState[Any]],
        halt_on: Optional[Callable[[InstanceState[Any]], bool]] = None,
    ) -> Iterator[
        Tuple[object, Mapper[Any], InstanceState[Any], _InstanceDict]
    ]:
        """Iterate through instances related to the given instance for
        a particular 'cascade', starting with this MapperProperty.

        Return an iterator3-tuples (instance, mapper, state).

        Note that the 'cascade' collection on this MapperProperty is
        checked first for the given type before cascade_iterator is called.

        This method typically only applies to Relationship.

        """

        return iter(())

    def set_parent(self, parent: Mapper[Any], init: bool) -> None:
        """Set the parent mapper that references this MapperProperty.

        This method is overridden by some subclasses to perform extra
        setup when the mapper is first known.

        """
        self.parent = parent

    def instrument_class(self, mapper: Mapper[Any]) -> None:
        """Hook called by the Mapper to the property to initiate
        instrumentation of the class attribute managed by this
        MapperProperty.

        The MapperProperty here will typically call out to the
        attributes module to set up an InstrumentedAttribute.

        This step is the first of two steps to set up an InstrumentedAttribute,
        and is called early in the mapper setup process.

        The second step is typically the init_class_attribute step,
        called from StrategizedProperty via the post_instrument_class()
        hook.  This step assigns additional state to the InstrumentedAttribute
        (specifically the "impl") which has been determined after the
        MapperProperty has determined what kind of persistence
        management it needs to do (e.g. scalar, object, collection, etc).

        """

    def __init__(
        self,
        attribute_options: Optional[_AttributeOptions] = None,
        _assume_readonly_dc_attributes: bool = False,
    ) -> None:
        self._configure_started = False
        self._configure_finished = False

        if _assume_readonly_dc_attributes:
            default_attrs = _DEFAULT_READONLY_ATTRIBUTE_OPTIONS
        else:
            default_attrs = _DEFAULT_ATTRIBUTE_OPTIONS

        if attribute_options and attribute_options != default_attrs:
            self._has_dataclass_arguments = True
            self._attribute_options = attribute_options
        else:
            self._has_dataclass_arguments = False
            self._attribute_options = default_attrs

    def init(self) -> None:
        """Called after all mappers are created to assemble
        relationships between mappers and perform other post-mapper-creation
        initialization steps.


        """
        self._configure_started = True
        self.do_init()
        self._configure_finished = True

    @property
    def class_attribute(self) -> InstrumentedAttribute[_T]:
        """Return the class-bound descriptor corresponding to this
        :class:`.MapperProperty`.

        This is basically a ``getattr()`` call::

            return getattr(self.parent.class_, self.key)

        I.e. if this :class:`.MapperProperty` were named ``addresses``,
        and the class to which it is mapped is ``User``, this sequence
        is possible::

            >>> from sqlalchemy import inspect
            >>> mapper = inspect(User)
            >>> addresses_property = mapper.attrs.addresses
            >>> addresses_property.class_attribute is User.addresses
            True
            >>> User.addresses.property is addresses_property
            True


        """

        return getattr(self.parent.class_, self.key)  # type: ignore

    def do_init(self) -> None:
        """Perform subclass-specific initialization post-mapper-creation
        steps.

        This is a template method called by the ``MapperProperty``
        object's init() method.

        """

    def post_instrument_class(self, mapper: Mapper[Any]) -> None:
        """Perform instrumentation adjustments that need to occur
        after init() has completed.

        The given Mapper is the Mapper invoking the operation, which
        may not be the same Mapper as self.parent in an inheritance
        scenario; however, Mapper will always at least be a sub-mapper of
        self.parent.

        This method is typically used by StrategizedProperty, which delegates
        it to LoaderStrategy.init_class_attribute() to perform final setup
        on the class-bound InstrumentedAttribute.

        """

    def merge(
        self,
        session: Session,
        source_state: InstanceState[Any],
        source_dict: _InstanceDict,
        dest_state: InstanceState[Any],
        dest_dict: _InstanceDict,
        load: bool,
        _recursive: Dict[Any, object],
        _resolve_conflict_map: Dict[_IdentityKeyType[Any], object],
    ) -> None:
        """Merge the attribute represented by this ``MapperProperty``
        from source to destination object.

        """

    def __repr__(self) -> str:
        return "<%s at 0x%x; %s>" % (
            self.__class__.__name__,
            id(self),
            getattr(self, "key", "no key"),
        )


@inspection._self_inspects
class PropComparator(SQLORMOperations[_T_co], Generic[_T_co], ColumnOperators):
    r"""Defines SQL operations for ORM mapped attributes.

    SQLAlchemy allows for operators to
    be redefined at both the Core and ORM level.  :class:`.PropComparator`
    is the base class of operator redefinition for ORM-level operations,
    including those of :class:`.ColumnProperty`,
    :class:`.Relationship`, and :class:`.Composite`.

    User-defined subclasses of :class:`.PropComparator` may be created. The
    built-in Python comparison and math operator methods, such as
    :meth:`.operators.ColumnOperators.__eq__`,
    :meth:`.operators.ColumnOperators.__lt__`, and
    :meth:`.operators.ColumnOperators.__add__`, can be overridden to provide
    new operator behavior. The custom :class:`.PropComparator` is passed to
    the :class:`.MapperProperty` instance via the ``comparator_factory``
    argument. In each case,
    the appropriate subclass of :class:`.PropComparator` should be used::

        # definition of custom PropComparator subclasses

        from sqlalchemy.orm.properties import \
                                ColumnProperty,\
                                Composite,\
                                Relationship

        class MyColumnComparator(ColumnProperty.Comparator):
            def __eq__(self, other):
                return self.__clause_element__() == other

        class MyRelationshipComparator(Relationship.Comparator):
            def any(self, expression):
                "define the 'any' operation"
                # ...

        class MyCompositeComparator(Composite.Comparator):
            def __gt__(self, other):
                "redefine the 'greater than' operation"

                return sql.and_(*[a>b for a, b in
                                  zip(self.__clause_element__().clauses,
                                      other.__composite_values__())])


        # application of custom PropComparator subclasses

        from sqlalchemy.orm import column_property, relationship, composite
        from sqlalchemy import Column, String

        class SomeMappedClass(Base):
            some_column = column_property(Column("some_column", String),
                                comparator_factory=MyColumnComparator)

            some_relationship = relationship(SomeOtherClass,
                                comparator_factory=MyRelationshipComparator)

            some_composite = composite(
                    Column("a", String), Column("b", String),
                    comparator_factory=MyCompositeComparator
                )

    Note that for column-level operator redefinition, it's usually
    simpler to define the operators at the Core level, using the
    :attr:`.TypeEngine.comparator_factory` attribute.  See
    :ref:`types_operators` for more detail.

    .. seealso::

        :class:`.ColumnProperty.Comparator`

        :class:`.Relationship.Comparator`

        :class:`.Composite.Comparator`

        :class:`.ColumnOperators`

        :ref:`types_operators`

        :attr:`.TypeEngine.comparator_factory`

    """
    __slots__ = "prop", "_parententity", "_adapt_to_entity"

    __visit_name__ = "orm_prop_comparator"

    _parententity: _InternalEntityType[Any]
    _adapt_to_entity: Optional[AliasedInsp[Any]]
    prop: RODescriptorReference[MapperProperty[_T_co]]

    def __init__(
        self,
        prop: MapperProperty[_T],
        parentmapper: _InternalEntityType[Any],
        adapt_to_entity: Optional[AliasedInsp[Any]] = None,
    ):
        self.prop = prop
        self._parententity = adapt_to_entity or parentmapper
        self._adapt_to_entity = adapt_to_entity

    @util.non_memoized_property
    def property(self) -> MapperProperty[_T]:
        """Return the :class:`.MapperProperty` associated with this
        :class:`.PropComparator`.


        Return values here will commonly be instances of
        :class:`.ColumnProperty` or :class:`.Relationship`.


        """
        return self.prop

    def __clause_element__(self) -> roles.ColumnsClauseRole:
        raise NotImplementedError("%r" % self)

    def _bulk_update_tuples(
        self, value: Any
    ) -> Sequence[Tuple[_DMLColumnArgument, Any]]:
        """Receive a SQL expression that represents a value in the SET
        clause of an UPDATE statement.

        Return a tuple that can be passed to a :class:`_expression.Update`
        construct.

        """

        return [(cast("_DMLColumnArgument", self.__clause_element__()), value)]

    def adapt_to_entity(
        self, adapt_to_entity: AliasedInsp[Any]
    ) -> PropComparator[_T]:
        """Return a copy of this PropComparator which will use the given
        :class:`.AliasedInsp` to produce corresponding expressions.
        """
        return self.__class__(self.prop, self._parententity, adapt_to_entity)

    @util.ro_non_memoized_property
    def _parentmapper(self) -> Mapper[Any]:
        """legacy; this is renamed to _parententity to be
        compatible with QueryableAttribute."""
        return self._parententity.mapper

    def _criterion_exists(
        self,
        criterion: Optional[_ColumnExpressionArgument[bool]] = None,
        **kwargs: Any,
    ) -> ColumnElement[Any]:
        return self.prop.comparator._criterion_exists(criterion, **kwargs)

    @util.ro_non_memoized_property
    def adapter(self) -> Optional[_ORMAdapterProto]:
        """Produce a callable that adapts column expressions
        to suit an aliased version of this comparator.

        """
        if self._adapt_to_entity is None:
            return None
        else:
            return self._adapt_to_entity._orm_adapt_element

    @util.ro_non_memoized_property
    def info(self) -> _InfoType:
        return self.prop.info

    @staticmethod
    def _any_op(a: Any, b: Any, **kwargs: Any) -> Any:
        return a.any(b, **kwargs)

    @staticmethod
    def _has_op(left: Any, other: Any, **kwargs: Any) -> Any:
        return left.has(other, **kwargs)

    @staticmethod
    def _of_type_op(a: Any, class_: Any) -> Any:
        return a.of_type(class_)

    any_op = cast(operators.OperatorType, _any_op)
    has_op = cast(operators.OperatorType, _has_op)
    of_type_op = cast(operators.OperatorType, _of_type_op)

    if typing.TYPE_CHECKING:

        def operate(
            self, op: OperatorType, *other: Any, **kwargs: Any
        ) -> ColumnElement[Any]:
            ...

        def reverse_operate(
            self, op: OperatorType, other: Any, **kwargs: Any
        ) -> ColumnElement[Any]:
            ...

    def of_type(self, class_: _EntityType[Any]) -> PropComparator[_T]:
        r"""Redefine this object in terms of a polymorphic subclass,
        :func:`_orm.with_polymorphic` construct, or :func:`_orm.aliased`
        construct.

        Returns a new PropComparator from which further criterion can be
        evaluated.

        e.g.::

            query.join(Company.employees.of_type(Engineer)).\
               filter(Engineer.name=='foo')

        :param \class_: a class or mapper indicating that criterion will be
            against this specific subclass.

        .. seealso::

            :ref:`orm_queryguide_joining_relationships_aliased` - in the
            :ref:`queryguide_toplevel`

            :ref:`inheritance_of_type`

        """

        return self.operate(PropComparator.of_type_op, class_)  # type: ignore

    def and_(
        self, *criteria: _ColumnExpressionArgument[bool]
    ) -> PropComparator[bool]:
        """Add additional criteria to the ON clause that's represented by this
        relationship attribute.

        E.g.::


            stmt = select(User).join(
                User.addresses.and_(Address.email_address != 'foo')
            )

            stmt = select(User).options(
                joinedload(User.addresses.and_(Address.email_address != 'foo'))
            )

        .. versionadded:: 1.4

        .. seealso::

            :ref:`orm_queryguide_join_on_augmented`

            :ref:`loader_option_criteria`

            :func:`.with_loader_criteria`

        """
        return self.operate(operators.and_, *criteria)  # type: ignore

    def any(
        self,
        criterion: Optional[_ColumnExpressionArgument[bool]] = None,
        **kwargs: Any,
    ) -> ColumnElement[bool]:
        r"""Return a SQL expression representing true if this element
        references a member which meets the given criterion.

        The usual implementation of ``any()`` is
        :meth:`.Relationship.Comparator.any`.

        :param criterion: an optional ClauseElement formulated against the
          member class' table or attributes.

        :param \**kwargs: key/value pairs corresponding to member class
          attribute names which will be compared via equality to the
          corresponding values.

        """

        return self.operate(PropComparator.any_op, criterion, **kwargs)

    def has(
        self,
        criterion: Optional[_ColumnExpressionArgument[bool]] = None,
        **kwargs: Any,
    ) -> ColumnElement[bool]:
        r"""Return a SQL expression representing true if this element
        references a member which meets the given criterion.

        The usual implementation of ``has()`` is
        :meth:`.Relationship.Comparator.has`.

        :param criterion: an optional ClauseElement formulated against the
          member class' table or attributes.

        :param \**kwargs: key/value pairs corresponding to member class
          attribute names which will be compared via equality to the
          corresponding values.

        """

        return self.operate(PropComparator.has_op, criterion, **kwargs)


class StrategizedProperty(MapperProperty[_T]):
    """A MapperProperty which uses selectable strategies to affect
    loading behavior.

    There is a single strategy selected by default.  Alternate
    strategies can be selected at Query time through the usage of
    ``StrategizedOption`` objects via the Query.options() method.

    The mechanics of StrategizedProperty are used for every Query
    invocation for every mapped attribute participating in that Query,
    to determine first how the attribute will be rendered in SQL
    and secondly how the attribute will retrieve a value from a result
    row and apply it to a mapped object.  The routines here are very
    performance-critical.

    """

    __slots__ = (
        "_strategies",
        "strategy",
        "_wildcard_token",
        "_default_path_loader_key",
        "strategy_key",
    )
    inherit_cache = True
    strategy_wildcard_key: ClassVar[str]

    strategy_key: _StrategyKey

    _strategies: Dict[_StrategyKey, LoaderStrategy]

    def _memoized_attr__wildcard_token(self) -> Tuple[str]:
        return (
            f"{self.strategy_wildcard_key}:{path_registry._WILDCARD_TOKEN}",
        )

    def _memoized_attr__default_path_loader_key(
        self,
    ) -> Tuple[str, Tuple[str]]:
        return (
            "loader",
            (f"{self.strategy_wildcard_key}:{path_registry._DEFAULT_TOKEN}",),
        )

    def _get_context_loader(
        self, context: ORMCompileState, path: AbstractEntityRegistry
    ) -> Optional[_LoadElement]:
        load: Optional[_LoadElement] = None

        search_path = path[self]

        # search among: exact match, "attr.*", "default" strategy
        # if any.
        for path_key in (
            search_path._loader_key,
            search_path._wildcard_path_loader_key,
            search_path._default_path_loader_key,
        ):
            if path_key in context.attributes:
                load = context.attributes[path_key]
                break

                # note that if strategy_options.Load is placing non-actionable
                # objects in the context like defaultload(), we would
                # need to continue the loop here if we got such an
                # option as below.
                # if load.strategy or load.local_opts:
                #    break

        return load

    def _get_strategy(self, key: _StrategyKey) -> LoaderStrategy:
        try:
            return self._strategies[key]
        except KeyError:
            pass

        # run outside to prevent transfer of exception context
        cls = self._strategy_lookup(self, *key)
        # this previously was setting self._strategies[cls], that's
        # a bad idea; should use strategy key at all times because every
        # strategy has multiple keys at this point
        self._strategies[key] = strategy = cls(self, key)
        return strategy

    def setup(
        self,
        context: ORMCompileState,
        query_entity: _MapperEntity,
        path: AbstractEntityRegistry,
        adapter: Optional[ORMAdapter],
        **kwargs: Any,
    ) -> None:
        loader = self._get_context_loader(context, path)
        if loader and loader.strategy:
            strat = self._get_strategy(loader.strategy)
        else:
            strat = self.strategy
        strat.setup_query(
            context, query_entity, path, loader, adapter, **kwargs
        )

    def create_row_processor(
        self,
        context: ORMCompileState,
        query_entity: _MapperEntity,
        path: AbstractEntityRegistry,
        mapper: Mapper[Any],
        result: Result[Any],
        adapter: Optional[ORMAdapter],
        populators: _PopulatorDict,
    ) -> None:
        loader = self._get_context_loader(context, path)
        if loader and loader.strategy:
            strat = self._get_strategy(loader.strategy)
        else:
            strat = self.strategy
        strat.create_row_processor(
            context,
            query_entity,
            path,
            loader,
            mapper,
            result,
            adapter,
            populators,
        )

    def do_init(self) -> None:
        self._strategies = {}
        self.strategy = self._get_strategy(self.strategy_key)

    def post_instrument_class(self, mapper: Mapper[Any]) -> None:
        if (
            not self.parent.non_primary
            and not mapper.class_manager._attr_has_impl(self.key)
        ):
            self.strategy.init_class_attribute(mapper)

    _all_strategies: collections.defaultdict[
        Type[MapperProperty[Any]], Dict[_StrategyKey, Type[LoaderStrategy]]
    ] = collections.defaultdict(dict)

    @classmethod
    def strategy_for(cls, **kw: Any) -> Callable[[_TLS], _TLS]:
        def decorate(dec_cls: _TLS) -> _TLS:
            # ensure each subclass of the strategy has its
            # own _strategy_keys collection
            if "_strategy_keys" not in dec_cls.__dict__:
                dec_cls._strategy_keys = []
            key = tuple(sorted(kw.items()))
            cls._all_strategies[cls][key] = dec_cls
            dec_cls._strategy_keys.append(key)
            return dec_cls

        return decorate

    @classmethod
    def _strategy_lookup(
        cls, requesting_property: MapperProperty[Any], *key: Any
    ) -> Type[LoaderStrategy]:
        requesting_property.parent._with_polymorphic_mappers

        for prop_cls in cls.__mro__:
            if prop_cls in cls._all_strategies:
                if TYPE_CHECKING:
                    assert issubclass(prop_cls, MapperProperty)
                strategies = cls._all_strategies[prop_cls]
                try:
                    return strategies[key]
                except KeyError:
                    pass

        for property_type, strats in cls._all_strategies.items():
            if key in strats:
                intended_property_type = property_type
                actual_strategy = strats[key]
                break
        else:
            intended_property_type = None
            actual_strategy = None

        raise orm_exc.LoaderStrategyException(
            cls,
            requesting_property,
            intended_property_type,
            actual_strategy,
            key,
        )


class ORMOption(ExecutableOption):
    """Base class for option objects that are passed to ORM queries.

    These options may be consumed by :meth:`.Query.options`,
    :meth:`.Select.options`, or in a more general sense by any
    :meth:`.Executable.options` method.   They are interpreted at
    statement compile time or execution time in modern use.  The
    deprecated :class:`.MapperOption` is consumed at ORM query construction
    time.

    .. versionadded:: 1.4

    """

    __slots__ = ()

    _is_legacy_option = False

    propagate_to_loaders = False
    """if True, indicate this option should be carried along
    to "secondary" SELECT statements that occur for relationship
    lazy loaders as well as attribute load / refresh operations.

    """

    _is_core = False

    _is_user_defined = False

    _is_compile_state = False

    _is_criteria_option = False

    _is_strategy_option = False

    def _adapt_cached_option_to_uncached_option(
        self, context: QueryContext, uncached_opt: ORMOption
    ) -> ORMOption:
        """adapt this option to the "uncached" version of itself in a
        loader strategy context.

        given "self" which is an option from a cached query, as well as the
        corresponding option from the uncached version of the same query,
        return the option we should use in a new query, in the context of a
        loader strategy being asked to load related rows on behalf of that
        cached query, which is assumed to be building a new query based on
        entities passed to us from the cached query.

        Currently this routine chooses between "self" and "uncached" without
        manufacturing anything new. If the option is itself a loader strategy
        option which has a path, that path needs to match to the entities being
        passed to us by the cached query, so the :class:`_orm.Load` subclass
        overrides this to return "self". For all other options, we return the
        uncached form which may have changing state, such as a
        with_loader_criteria() option which will very often have new state.

        This routine could in the future involve
        generating a new option based on both inputs if use cases arise,
        such as if with_loader_criteria() needed to match up to
        ``AliasedClass`` instances given in the parent query.

        However, longer term it might be better to restructure things such that
        ``AliasedClass`` entities are always matched up on their cache key,
        instead of identity, in things like paths and such, so that this whole
        issue of "the uncached option does not match the entities" goes away.
        However this would make ``PathRegistry`` more complicated and difficult
        to debug as well as potentially less performant in that it would be
        hashing enormous cache keys rather than a simple AliasedInsp. UNLESS,
        we could get cache keys overall to be reliably hashed into something
        like an md5 key.

        .. versionadded:: 1.4.41

        """
        if uncached_opt is not None:
            return uncached_opt
        else:
            return self


class CompileStateOption(HasCacheKey, ORMOption):
    """base for :class:`.ORMOption` classes that affect the compilation of
    a SQL query and therefore need to be part of the cache key.

    .. note::  :class:`.CompileStateOption` is generally non-public and
       should not be used as a base class for user-defined options; instead,
       use :class:`.UserDefinedOption`, which is easier to use as it does not
       interact with ORM compilation internals or caching.

    :class:`.CompileStateOption` defines an internal attribute
    ``_is_compile_state=True`` which has the effect of the ORM compilation
    routines for SELECT and other statements will call upon these options when
    a SQL string is being compiled. As such, these classes implement
    :class:`.HasCacheKey` and need to provide robust ``_cache_key_traversal``
    structures.

    The :class:`.CompileStateOption` class is used to implement the ORM
    :class:`.LoaderOption` and :class:`.CriteriaOption` classes.

    .. versionadded:: 1.4.28


    """

    __slots__ = ()

    _is_compile_state = True

    def process_compile_state(self, compile_state: ORMCompileState) -> None:
        """Apply a modification to a given :class:`.ORMCompileState`.

        This method is part of the implementation of a particular
        :class:`.CompileStateOption` and is only invoked internally
        when an ORM query is compiled.

        """

    def process_compile_state_replaced_entities(
        self,
        compile_state: ORMCompileState,
        mapper_entities: Sequence[_MapperEntity],
    ) -> None:
        """Apply a modification to a given :class:`.ORMCompileState`,
        given entities that were replaced by with_only_columns() or
        with_entities().

        This method is part of the implementation of a particular
        :class:`.CompileStateOption` and is only invoked internally
        when an ORM query is compiled.

        .. versionadded:: 1.4.19

        """


class LoaderOption(CompileStateOption):
    """Describe a loader modification to an ORM statement at compilation time.

    .. versionadded:: 1.4

    """

    __slots__ = ()

    def process_compile_state_replaced_entities(
        self,
        compile_state: ORMCompileState,
        mapper_entities: Sequence[_MapperEntity],
    ) -> None:
        self.process_compile_state(compile_state)


class CriteriaOption(CompileStateOption):
    """Describe a WHERE criteria modification to an ORM statement at
    compilation time.

    .. versionadded:: 1.4

    """

    __slots__ = ()

    _is_criteria_option = True

    def get_global_criteria(self, attributes: Dict[str, Any]) -> None:
        """update additional entity criteria options in the given
        attributes dictionary.

        """


class UserDefinedOption(ORMOption):
    """Base class for a user-defined option that can be consumed from the
    :meth:`.SessionEvents.do_orm_execute` event hook.

    """

    __slots__ = ("payload",)

    _is_legacy_option = False

    _is_user_defined = True

    propagate_to_loaders = False
    """if True, indicate this option should be carried along
    to "secondary" Query objects produced during lazy loads
    or refresh operations.

    """

    def __init__(self, payload: Optional[Any] = None):
        self.payload = payload


@util.deprecated_cls(
    "1.4",
    "The :class:`.MapperOption class is deprecated and will be removed "
    "in a future release.   For "
    "modifications to queries on a per-execution basis, use the "
    ":class:`.UserDefinedOption` class to establish state within a "
    ":class:`.Query` or other Core statement, then use the "
    ":meth:`.SessionEvents.before_orm_execute` hook to consume them.",
    constructor=None,
)
class MapperOption(ORMOption):
    """Describe a modification to a Query"""

    __slots__ = ()

    _is_legacy_option = True

    propagate_to_loaders = False
    """if True, indicate this option should be carried along
    to "secondary" Query objects produced during lazy loads
    or refresh operations.

    """

    def process_query(self, query: Query[Any]) -> None:
        """Apply a modification to the given :class:`_query.Query`."""

    def process_query_conditionally(self, query: Query[Any]) -> None:
        """same as process_query(), except that this option may not
        apply to the given query.

        This is typically applied during a lazy load or scalar refresh
        operation to propagate options stated in the original Query to the
        new Query being used for the load.  It occurs for those options that
        specify propagate_to_loaders=True.

        """

        self.process_query(query)


class LoaderStrategy:
    """Describe the loading behavior of a StrategizedProperty object.

    The ``LoaderStrategy`` interacts with the querying process in three
    ways:

    * it controls the configuration of the ``InstrumentedAttribute``
      placed on a class to handle the behavior of the attribute.  this
      may involve setting up class-level callable functions to fire
      off a select operation when the attribute is first accessed
      (i.e. a lazy load)

    * it processes the ``QueryContext`` at statement construction time,
      where it can modify the SQL statement that is being produced.
      For example, simple column attributes will add their represented
      column to the list of selected columns, a joined eager loader
      may establish join clauses to add to the statement.

    * It produces "row processor" functions at result fetching time.
      These "row processor" functions populate a particular attribute
      on a particular mapped instance.

    """

    __slots__ = (
        "parent_property",
        "is_class_level",
        "parent",
        "key",
        "strategy_key",
        "strategy_opts",
    )

    _strategy_keys: ClassVar[List[_StrategyKey]]

    def __init__(
        self, parent: MapperProperty[Any], strategy_key: _StrategyKey
    ):
        self.parent_property = parent
        self.is_class_level = False
        self.parent = self.parent_property.parent
        self.key = self.parent_property.key
        self.strategy_key = strategy_key
        self.strategy_opts = dict(strategy_key)

    def init_class_attribute(self, mapper: Mapper[Any]) -> None:
        pass

    def setup_query(
        self,
        compile_state: ORMCompileState,
        query_entity: _MapperEntity,
        path: AbstractEntityRegistry,
        loadopt: Optional[_LoadElement],
        adapter: Optional[ORMAdapter],
        **kwargs: Any,
    ) -> None:
        """Establish column and other state for a given QueryContext.

        This method fulfills the contract specified by MapperProperty.setup().

        StrategizedProperty delegates its setup() method
        directly to this method.

        """

    def create_row_processor(
        self,
        context: ORMCompileState,
        query_entity: _MapperEntity,
        path: AbstractEntityRegistry,
        loadopt: Optional[_LoadElement],
        mapper: Mapper[Any],
        result: Result[Any],
        adapter: Optional[ORMAdapter],
        populators: _PopulatorDict,
    ) -> None:
        """Establish row processing functions for a given QueryContext.

        This method fulfills the contract specified by
        MapperProperty.create_row_processor().

        StrategizedProperty delegates its create_row_processor() method
        directly to this method.

        """

    def __str__(self) -> str:
        return str(self.parent_property)



import json

from typing import (
    Dict,
    List,
    Sequence,
)

from .error import UpdateItemsError
from .shared.dto import RuptureEventDto
from .shared.dto.exception import InvalidPartCode
from .token import (
    TK_SLCTRL_PMGR_GET_TENDERS,
    TK_SLCTRL_PMGR_GET_PRODUCTS_NAME,
    TK_SLCTRL_PMGR_RUPTURE_UPDATE_DAILY_DISHES,
    TK_SLCTRL_PMGR_RUPTURE_UPDATE_ITEMS,
)

from messagebus import (
    ApplicationErrorMessage,
    MessageBus,
    Message,
    DataType,
    DefaultToken,
)


class ProductService(object):

    def __init__(
        self,               # type: ProductService
        message_bus,        # type: MessageBus
    ):
        # type: (...) -> None

        self.message_bus = message_bus
        self.component_name = "ORDERMGR"

    def get_tenders_list(
        self,       # type: ProductService
    ):
        # type: (...) -> Dict

        msg = self._get_tenders()
        json_data = json.loads(msg.data)
        return _byteify(json_data)

    def get_tenders_dict(
        self,       # type: ProductService
    ):
        # type: (...) -> Dict

        msg = self._get_tenders()
        tenders = json.loads(msg.data)
        tenders_dict = {}
        for tender in tenders:
            tenders_dict[tender["id"]] = tender

        return _byteify(tenders_dict)

    def get_products_name(
        self,   # type: ProductService
    ):
        # type: (...) -> Dict[str, str]

        reply = self.message_bus.send_message(
            component=self.component_name,
            message=Message(token=TK_SLCTRL_PMGR_GET_PRODUCTS_NAME)
        )

        if reply.token != DefaultToken.TK_SYS_ACK:
            raise RuntimeError("Error getting products name")

        return json.loads(reply.data.decode())

    def update_items(
        self,           # type: ProductService
        user_id,        # type: int
        enabled,        # type: Sequence[int]
        disabled,       # type: Sequence[int]
    ):
        # type: (...) -> RuptureEventDto

        reply = self.message_bus.send_message(
            component=self.component_name,
            message=Message(
                token=TK_SLCTRL_PMGR_RUPTURE_UPDATE_ITEMS,
                data=json.dumps({
                    "user_id": user_id,
                    "enabled": enabled,
                    "disabled": disabled,
                }).encode()
            )
        )

        if reply.token == DefaultToken.TK_SYS_APPLICATION_ERROR:
            application_error = ApplicationErrorMessage.build_from_message(message=reply)
            if application_error.error_code == UpdateItemsError.INVALID_PART_CODE.code:
                raise InvalidPartCode(part_code=int(application_error.error_message))

        data_dict = json.loads(reply.data.decode())
        return RuptureEventDto.build_from_dict(data_dict)

    def update_daily_dishes(
        self,       # type: ProductService
        user_id,    # type: int
        to_add,     # type: List[Dict[str, int]]
        to_remove,  # type: List[int]
    ):
        # type: (...) -> None

        dishes_changes_json = json.dumps({
            "userId": user_id,
            "unset": to_remove,
            "set": to_add,
        }).encode()

        reply = self.message_bus.send_message(
            component=self.component_name,
            message=Message(
                token=TK_SLCTRL_PMGR_RUPTURE_UPDATE_DAILY_DISHES,
                data=dishes_changes_json,
            ),
        )

        if reply.token != DefaultToken.TK_SYS_ACK:
            raise RuntimeError("Could not update daily dishes")

        return None

    def _get_tenders(
        self,       # type: ProductService
    ):
        # type: (...) -> Message

        msg = self.message_bus.send_message(
            component=self.component_name,
            message=Message(
                token=TK_SLCTRL_PMGR_GET_TENDERS,
                data_type=DataType.param,
            )
        )

        if msg.token != DefaultToken.TK_SYS_ACK:
            raise RuntimeError("Could not retrieve tenders")

        return msg


def _byteify(input):
    if isinstance(input, dict):
        return {_byteify(key): _byteify(value) for key, value in input.iteritems()}

    if isinstance(input, list):
        return [_byteify(element) for element in input]

    if isinstance(input, str):
        return input.encode()

    return input



import sys

if sys.version_info[0] == 2:
    from . import _compat_py2 as _compat_impl
else:
    from . import _compat_py3 as _compat_impl

UTC = _compat_impl.UTC
get_timezone = _compat_impl.get_timezone
get_timezone_file = _compat_impl.get_timezone_file
get_fixed_offset_zone = _compat_impl.get_fixed_offset_zone
is_ambiguous = _compat_impl.is_ambiguous
is_imaginary = _compat_impl.is_imaginary
enfold = _compat_impl.enfold
get_fold = _compat_impl.get_fold



import pyscripts

from mbcontextmessagehandler import MbContextMessageBus
from product_api import ProductService


def get_product_service():
    # type: (...) -> ProductService

    return ProductService(
        message_bus=MbContextMessageBus(mbcontext=pyscripts.mbcontext),
    )



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from functools import lru_cache

from s3transfer.compat import accepts_kwargs
from s3transfer.exceptions import InvalidSubscriberMethodError


class BaseSubscriber:
    """The base subscriber class

    It is recommended that all subscriber implementations subclass and then
    override the subscription methods (i.e. on_{subsribe_type}() methods).
    """

    VALID_SUBSCRIBER_TYPES = ['queued', 'progress', 'done']

    def __new__(cls, *args, **kwargs):
        cls._validate_subscriber_methods()
        return super().__new__(cls)

    @classmethod
    @lru_cache()
    def _validate_subscriber_methods(cls):
        for subscriber_type in cls.VALID_SUBSCRIBER_TYPES:
            subscriber_method = getattr(cls, 'on_' + subscriber_type)
            if not callable(subscriber_method):
                raise InvalidSubscriberMethodError(
                    'Subscriber method %s must be callable.'
                    % subscriber_method
                )

            if not accepts_kwargs(subscriber_method):
                raise InvalidSubscriberMethodError(
                    'Subscriber method %s must accept keyword '
                    'arguments (**kwargs)' % subscriber_method
                )

    def on_queued(self, future, **kwargs):
        """Callback to be invoked when transfer request gets queued

        This callback can be useful for:

            * Keeping track of how many transfers have been requested
            * Providing the expected transfer size through
              future.meta.provide_transfer_size() so a HeadObject would not
              need to be made for copies and downloads.

        :type future: s3transfer.futures.TransferFuture
        :param future: The TransferFuture representing the requested transfer.
        """
        pass

    def on_progress(self, future, bytes_transferred, **kwargs):
        """Callback to be invoked when progress is made on transfer

        This callback can be useful for:

            * Recording and displaying progress

        :type future: s3transfer.futures.TransferFuture
        :param future: The TransferFuture representing the requested transfer.

        :type bytes_transferred: int
        :param bytes_transferred: The number of bytes transferred for that
            invocation of the callback. Note that a negative amount can be
            provided, which usually indicates that an in-progress request
            needed to be retried and thus progress was rewound.
        """
        pass

    def on_done(self, future, **kwargs):
        """Callback to be invoked once a transfer is done

        This callback can be useful for:

            * Recording and displaying whether the transfer succeeded or
              failed using future.result()
            * Running some task after the transfer completed like changing
              the last modified time of a downloaded file.

        :type future: s3transfer.futures.TransferFuture
        :param future: The TransferFuture representing the requested transfer.
        """
        pass



from datetime import datetime
from logging import getLogger
from threading import (
    Thread,
    Condition,
)

from production.repository import ProductionRepository
from production.model import PicklistCleanerManagerConfig

logger = getLogger("PicklistDatabaseManager")


class PicklistCleanerManager(object):
    def __init__(
        self,                   # type: PicklistCleanerManager
        production_repository,  # type: ProductionRepository
        config,                 # type: PicklistCleanerManagerConfig
    ):
        # type: (...) -> None

        super(PicklistCleanerManager, self).__init__()
        self.production_repository = production_repository
        self.start_time = config.start_time
        self.end_time = config.end_time
        self.days_to_keep_data = config.days_to_keep_data

        self.condition = Condition()
        self.running = False
        self.minutes_to_call = 30
        self.has_already_applied_cleaning = False

    def handle_clean_databases(
        self,   # type: PicklistCleanerManager
    ):
        # type: (...) -> None

        logger.info("Start cleaning picklist databases")

        try:
            self.production_repository.delete_outdated_pick_list_orders(range_days=self.days_to_keep_data)
            vacuum_size_info = self.production_repository.vacuum_database()
            self.has_already_applied_cleaning = True

            log_message = "Production db file sizes. OLD: {} bytes; NEW: {} bytes".format(
                vacuum_size_info.get("old_file_size"),
                vacuum_size_info.get("new_file_size"),
            )
            logger.info(log_message)
        except Exception as _:
            logger.exception(
                "Error cleaning picklist databases",
                exc_info=True,
            )

        logger.info("Picklist databases cleaned with success")

    def start(
        self,   # type: PicklistCleanerManager
    ):
        # type: (...) -> None

        if not self.running:
            self.running = True
            thread = Thread(target=self._cleaner)
            thread.name = "PicklistDatabaseManager"
            thread.daemon = True
            thread.start()

    def stop(
        self,   # type: PicklistCleanerManager
    ):
        # type: (...) -> None

        if self.running:
            with self.condition:
                self.running = False
                self.condition.notify_all()

    def _cleaner(
        self,   # type: PicklistCleanerManager
    ):
        # type: (...) -> None

        self.handle_clean_databases()

        if self._clean_only_on_component_start():
            self.stop()
            return

        while self.running:
            if self.is_in_maintenance_window():
                self.handle_clean_databases()
            self._sleep()

    def _clean_only_on_component_start(
        self,   # type: PicklistCleanerManager
    ):
        # type: (...) -> bool

        return self.start_time is None or self.end_time is None

    def _sleep(
        self,  # type: PicklistCleanerManager
    ):
        # type: (...) -> None

        with self.condition:
            self.condition.wait(self.minutes_to_call * 60)

    def is_in_maintenance_window(
        self,        # type: PicklistCleanerManager
    ):
        # type: (...) -> bool

        now = datetime.now().time()
        is_inside_clean_window = self.start_time <= now <= self.end_time

        if self.has_already_applied_cleaning and not is_inside_clean_window:
            self.has_already_applied_cleaning = False

        if not self.has_already_applied_cleaning and is_inside_clean_window:
            return True

        return False




import time

from abc import (
    ABCMeta,
    abstractmethod,
)
from logging import Logger
from threading import Thread
from typing import List

from mw_helper import ensure_list

from production.model import (
    ProductionOrder,
    InvalidOrderToPrint,
)
from production.view import ProductionView


class PrinterView(ProductionView):
    __metaclass__ = ABCMeta

    def __init__(
        self,
        name,                   # type: str
        tags_on_print=None,     # type: List[str]
        logger=None,            # type: Logger
    ):
        # type: (...) -> None

        super(PrinterView, self).__init__(name, logger)
        self.tags_on_print = ensure_list(tags_on_print)

    def handle_order(
        self,               # type: PrinterView
        order,              # type: ProductionOrder
        synch=False,        # type: bool
    ):
        # type: (...) -> Thread

        t = Thread(
            target=self.generate_and_print_report,
            args=[order]
        )
        t.daemon = True
        t.start()
        return t

    def generate_and_print_report(
        self,       # type: PrinterView
        order,      # type: ProductionOrder
    ):
        # type: (...) -> None

        report = self.generate_report_with_retry(order)
        self.print_report(report, order)

    def generate_report_with_retry(
        self,       # type: PrinterView
        order,      # type: ProductionOrder
    ):
        # type: (...) -> str

        report = None
        for _ in range(1, 5):
            try:
                report = self.generate_report(order)
                break
            except InvalidOrderToPrint:
                break
            except (Exception,):
                self.exception("Exception trying to generate picklist for Order {}", order.order_id)
                time.sleep(1)

        return report

    @abstractmethod
    def generate_report(
        self,       # type: PrinterView
        order,      # type: ProductionOrder
    ):
        # type: (...) -> str

        raise NotImplementedError()

    @abstractmethod
    def print_report(
        self,       # type: PrinterView
        report,     # type: str
        order,      # type: ProductionOrder
    ):
        # type: (...) -> None

        raise NotImplementedError()

    @abstractmethod
    def clear_printer_cache(
        self,       # type: PrinterView
    ):
        # type: (...) -> None

        raise NotImplementedError()

    def get_view_tags(
        self,       # type: PrinterView
    ):
        # type: (...) -> List[str]

        return self.tags_on_print



# -*- coding: utf-8 -*-

from typing import Optional
from xml.etree import cElementTree as eTree

from custom_sysactions import block_action_if_pos_is_blocked
from sysactions import (
    StopAction,
    action,
    get_model,
    get_posot,
    show_messagebox,
    show_keyboard,
    send_message,
    close_asynch_dialog,
)
from msgbus import (
    TK_SYS_ACK,
    FM_PARAM,
    MBException,
    TK_DRV_IOCTL,
)
from systools import sys_log_exception

from peripherals.scale import (
    PosScaleEvents,
    ScaleEvent,
)
from application.interactor import OrderInteractor
from peripherals.scale import WeightComparer
from .drive_thru import validate_can_change_order
from utilfunctions import get_device_service


waiting_scale = False
pos_scale_events = PosScaleEvents()


@action
def doCompleteOption(
    pos_id,                # type: str
    context,               # type: str
    part_code,             # type: str
    qty="1",               # type: str
    line_number="",        # type: str
    sale_type="EAT_IN",    # type: str
    subst="",              # type: str
    seat="",               # type: str
    service_period="",     # type: str
    pricelist="",          # type: str
    comment=None,          # type: Optional[str]
    *args
):
    # type: (...) -> None

    from posactions import (
        is_table_service,
        is_a_product_with_measure_unit,
        is_changing_order,
    )

    global waiting_scale

    model = get_model(posid=pos_id)
    is_order_table_service = is_table_service(model=model)
    if is_order_table_service and is_changing_order(model=model):
        show_messagebox(
            posid=pos_id,
            message="$CANNOT_ADD_ITEMS_IN_AN_OPENED_ORDER",
            icon="error",
        )
        raise StopAction()

    order_interactor = OrderInteractor(pos_id=pos_id)
    if not is_order_table_service or not order_interactor.is_order_in_progress():
        _verify_pos_is_blocked(pos_id=pos_id)

    order_picture = order_interactor.get_order_picture(pos_id=pos_id)
    sale_type = order_picture.get("saleTypeDescr") if not sale_type and order_picture else sale_type
    if order_picture and order_picture.get("state") == "IN_PROGRESS":
        validate_can_change_order(
            pos_id=pos_id,
            order_picture=order_picture,
            line_to_replace=subst,
            context=context,
            line_number=line_number,
        )

    if is_a_product_with_measure_unit(product_code=part_code) and not isinstance(qty, float):
        if waiting_scale:
            return

        try:
            waiting_scale = True
            qty = _get_weight_qty(pos_id=pos_id)
        finally:
            waiting_scale = False

    sale_type = sale_type or "EAT_IN"

    return _do_complete_option(
        pos_id=pos_id,
        model=model,
        subst=subst,
        line_number=line_number,
        context=context,
        part_code=part_code,
        sale_type=sale_type,
        comment=comment,
        price_list=pricelist,
        service_period=service_period,
        qty=qty,
        seat=seat,
        *args
    )


def _do_complete_option(
    pos_id,            # type: str
    model,             # type: eTree.XML
    subst,             # type: str
    line_number,       # type: str
    context,           # type: str
    part_code,         # type: str
    sale_type,         # type: str
    comment,           # type: str
    price_list,        # type: str
    service_period,    # type: str
    qty,               # type: str
    seat,              # type: str
    *args
):
    # type: (...) -> None

    from posactions import (
        doClearOptionItem,
        doOption,
    )

    sale_xml = None
    pos_ot = get_posot(model=model)
    try:
        if subst:
            if qty != "0":
                pos_ot.blkopnotify = True

            doClearOptionItem(
                pos_id=pos_id,
                linenumber=line_number,
                item_id=subst,
                pos_ot=pos_ot,
            )
            pos_ot.blkopnotify = False

        if qty != "0":
            sale_xml = doOption(
                pos_id=pos_id,
                context=context,
                part_code=part_code,
                qty=qty,
                line_number=line_number,
                sale_type=sale_type,
                pricelist=price_list,
                comment=comment,
                pos_ot=pos_ot,
                seat=seat,
                service_period=service_period,
                *args
            )

        if sale_xml is None:
            return

        return sale_xml
    finally:
        pos_ot.blkopnotify = False


@block_action_if_pos_is_blocked
def _verify_pos_is_blocked(
    pos_id,    # type: str
):
    # type: (...) -> None

    return


def _get_weight_qty(
    pos_id,    # type: str
):
    # type: (...) -> str

    weight_qty = _get_scale_weight(pos_id=pos_id)
    if weight_qty is None:
        weight_qty = show_keyboard(
            posid=pos_id,
            message="$TYPE_WEIGHT_QTY",
            title="",
            mask="weight",
            defvalue="0.000",
            numpad=True,
        )
    if weight_qty in (None, "", "."):
        raise StopAction()

    if weight_qty == "0":
        raise StopAction()

    return str(float(weight_qty))


def _get_scale_weight(
    pos_id,    # type: str
):
    # type: (...) -> str

    scale_name = get_device_service(
        pos_id=pos_id,
        device_type="scale",
    )
    weight = _scale_wait_data(
        pos_id=pos_id,
        service_name=scale_name,
    )
    return weight


def _start_measuring_weight(
    service_name,    # type: str
):
    # type: (...) -> bool

    try:
        msg = send_message(
            dest_name=service_name,
            token=TK_DRV_IOCTL,
            format=FM_PARAM,
            data="",
        )
        if msg.token == TK_SYS_ACK:
            return True
        else:
            return False

    except (MBException,):
        sys_log_exception("Error communicating with the scale")
        return False


def _scale_wait_data(
    pos_id,          # type: str
    service_name,    # type: str
    timeout=2000,    # type: Optional[float]
):
    # type: (...) -> Optional[str]

    scale_event = pos_scale_events.create_scale_event(
        pos_id=pos_id,
        service_name=service_name,
    )

    return _get_scaled_data_weight(
        pos_id=pos_id,
        service_name=service_name,
        scale_event=scale_event,
        timeout=timeout,
    )


def _get_scaled_data_weight(
    pos_id,          # type: str
    service_name,    # type: str
    scale_event,     # type: ScaleEvent
    timeout,         # type: float
):
    # type: (...) -> Optional[str]

    from posactions import pos_scale_dialogs

    dialog_id = None
    weight_comparer = WeightComparer()
    try:
        while True:
            if not _start_measuring_weight(service_name=service_name):
                return None

            if dialog_id is None:
                dialog_id = show_messagebox(
                    posid=pos_id,
                    title="$WAIT_SCALE_WEIGHT",
                    buttons="$CANCEL",
                    asynch=True,
                )
                pos_scale_dialogs.append(str(dialog_id))
            elif str(dialog_id) not in pos_scale_dialogs:
                return ""

            scale_event.wait(float(timeout / 1000.0))
            if scale_event.weight is None:
                return None

            if scale_event.weight > 0 and weight_comparer.is_weight_equal_to_previous(weight=scale_event.weight):
                close_asynch_dialog(
                    posid=pos_id,
                    dlgid=dialog_id,
                )
                dialog_id = None

                selected_option = show_messagebox(
                    posid=pos_id,
                    message="$CONFIRM_SCALE_WEIGHT|{0}".format("%.3f" % scale_event.weight),
                    buttons="$OK|$GET_WEIGHT_AGAIN|$CANCEL",
                )

                if selected_option == 0:
                    return scale_event.weight
                elif selected_option == 2:
                    return ""
    finally:
        _update_scale_event(
            pos_id=pos_id,
            dialog_id=dialog_id,
            scale_event=scale_event,
        )


def _update_scale_event(
    pos_id,         # type: str
    dialog_id,      # type: str
    scale_event,    # type: ScaleEvent
):
    # type: (...) -> None

    if pos_id in pos_scale_events.get_wait_pos_id_event_dict():
        pos_scale_events.finish_event(scale_event=scale_event)

    if dialog_id is not None:
        close_asynch_dialog(
            posid=pos_id,
            dlgid=dialog_id,
        )



from ._CatalogUpdaterAPIRepository import CatalogUpdaterAPIRepository
from ._MediaUpdaterAPIRepository import MediaUpdaterAPIRepository
from ._UserUpdaterAPIRepository import UserUpdaterAPIRepository
from ._LoaderUpdaterAPIRepository import LoaderUpdaterAPIRepository
from ._BinaryUpdaterAPIRepository import BinaryUpdaterAPIRepository



data = (
'hwen',    # 0x00
'hwenj',    # 0x01
'hwenh',    # 0x02
'hwed',    # 0x03
'hwel',    # 0x04
'hwelg',    # 0x05
'hwelm',    # 0x06
'hwelb',    # 0x07
'hwels',    # 0x08
'hwelt',    # 0x09
'hwelp',    # 0x0a
'hwelh',    # 0x0b
'hwem',    # 0x0c
'hweb',    # 0x0d
'hwebs',    # 0x0e
'hwes',    # 0x0f
'hwess',    # 0x10
'hweng',    # 0x11
'hwej',    # 0x12
'hwec',    # 0x13
'hwek',    # 0x14
'hwet',    # 0x15
'hwep',    # 0x16
'hweh',    # 0x17
'hwi',    # 0x18
'hwig',    # 0x19
'hwigg',    # 0x1a
'hwigs',    # 0x1b
'hwin',    # 0x1c
'hwinj',    # 0x1d
'hwinh',    # 0x1e
'hwid',    # 0x1f
'hwil',    # 0x20
'hwilg',    # 0x21
'hwilm',    # 0x22
'hwilb',    # 0x23
'hwils',    # 0x24
'hwilt',    # 0x25
'hwilp',    # 0x26
'hwilh',    # 0x27
'hwim',    # 0x28
'hwib',    # 0x29
'hwibs',    # 0x2a
'hwis',    # 0x2b
'hwiss',    # 0x2c
'hwing',    # 0x2d
'hwij',    # 0x2e
'hwic',    # 0x2f
'hwik',    # 0x30
'hwit',    # 0x31
'hwip',    # 0x32
'hwih',    # 0x33
'hyu',    # 0x34
'hyug',    # 0x35
'hyugg',    # 0x36
'hyugs',    # 0x37
'hyun',    # 0x38
'hyunj',    # 0x39
'hyunh',    # 0x3a
'hyud',    # 0x3b
'hyul',    # 0x3c
'hyulg',    # 0x3d
'hyulm',    # 0x3e
'hyulb',    # 0x3f
'hyuls',    # 0x40
'hyult',    # 0x41
'hyulp',    # 0x42
'hyulh',    # 0x43
'hyum',    # 0x44
'hyub',    # 0x45
'hyubs',    # 0x46
'hyus',    # 0x47
'hyuss',    # 0x48
'hyung',    # 0x49
'hyuj',    # 0x4a
'hyuc',    # 0x4b
'hyuk',    # 0x4c
'hyut',    # 0x4d
'hyup',    # 0x4e
'hyuh',    # 0x4f
'heu',    # 0x50
'heug',    # 0x51
'heugg',    # 0x52
'heugs',    # 0x53
'heun',    # 0x54
'heunj',    # 0x55
'heunh',    # 0x56
'heud',    # 0x57
'heul',    # 0x58
'heulg',    # 0x59
'heulm',    # 0x5a
'heulb',    # 0x5b
'heuls',    # 0x5c
'heult',    # 0x5d
'heulp',    # 0x5e
'heulh',    # 0x5f
'heum',    # 0x60
'heub',    # 0x61
'heubs',    # 0x62
'heus',    # 0x63
'heuss',    # 0x64
'heung',    # 0x65
'heuj',    # 0x66
'heuc',    # 0x67
'heuk',    # 0x68
'heut',    # 0x69
'heup',    # 0x6a
'heuh',    # 0x6b
'hyi',    # 0x6c
'hyig',    # 0x6d
'hyigg',    # 0x6e
'hyigs',    # 0x6f
'hyin',    # 0x70
'hyinj',    # 0x71
'hyinh',    # 0x72
'hyid',    # 0x73
'hyil',    # 0x74
'hyilg',    # 0x75
'hyilm',    # 0x76
'hyilb',    # 0x77
'hyils',    # 0x78
'hyilt',    # 0x79
'hyilp',    # 0x7a
'hyilh',    # 0x7b
'hyim',    # 0x7c
'hyib',    # 0x7d
'hyibs',    # 0x7e
'hyis',    # 0x7f
'hyiss',    # 0x80
'hying',    # 0x81
'hyij',    # 0x82
'hyic',    # 0x83
'hyik',    # 0x84
'hyit',    # 0x85
'hyip',    # 0x86
'hyih',    # 0x87
'hi',    # 0x88
'hig',    # 0x89
'higg',    # 0x8a
'higs',    # 0x8b
'hin',    # 0x8c
'hinj',    # 0x8d
'hinh',    # 0x8e
'hid',    # 0x8f
'hil',    # 0x90
'hilg',    # 0x91
'hilm',    # 0x92
'hilb',    # 0x93
'hils',    # 0x94
'hilt',    # 0x95
'hilp',    # 0x96
'hilh',    # 0x97
'him',    # 0x98
'hib',    # 0x99
'hibs',    # 0x9a
'his',    # 0x9b
'hiss',    # 0x9c
'hing',    # 0x9d
'hij',    # 0x9e
'hic',    # 0x9f
'hik',    # 0xa0
'hit',    # 0xa1
'hip',    # 0xa2
'hih',    # 0xa3
None,    # 0xa4
None,    # 0xa5
None,    # 0xa6
None,    # 0xa7
None,    # 0xa8
None,    # 0xa9
None,    # 0xaa
None,    # 0xab
None,    # 0xac
None,    # 0xad
None,    # 0xae
None,    # 0xaf
None,    # 0xb0
None,    # 0xb1
None,    # 0xb2
None,    # 0xb3
None,    # 0xb4
None,    # 0xb5
None,    # 0xb6
None,    # 0xb7
None,    # 0xb8
None,    # 0xb9
None,    # 0xba
None,    # 0xbb
None,    # 0xbc
None,    # 0xbd
None,    # 0xbe
None,    # 0xbf
None,    # 0xc0
None,    # 0xc1
None,    # 0xc2
None,    # 0xc3
None,    # 0xc4
None,    # 0xc5
None,    # 0xc6
None,    # 0xc7
None,    # 0xc8
None,    # 0xc9
None,    # 0xca
None,    # 0xcb
None,    # 0xcc
None,    # 0xcd
None,    # 0xce
None,    # 0xcf
None,    # 0xd0
None,    # 0xd1
None,    # 0xd2
None,    # 0xd3
None,    # 0xd4
None,    # 0xd5
None,    # 0xd6
None,    # 0xd7
None,    # 0xd8
None,    # 0xd9
None,    # 0xda
None,    # 0xdb
None,    # 0xdc
None,    # 0xdd
None,    # 0xde
None,    # 0xdf
None,    # 0xe0
None,    # 0xe1
None,    # 0xe2
None,    # 0xe3
None,    # 0xe4
None,    # 0xe5
None,    # 0xe6
None,    # 0xe7
None,    # 0xe8
None,    # 0xe9
None,    # 0xea
None,    # 0xeb
None,    # 0xec
None,    # 0xed
None,    # 0xee
None,    # 0xef
None,    # 0xf0
None,    # 0xf1
None,    # 0xf2
None,    # 0xf3
None,    # 0xf4
None,    # 0xf5
None,    # 0xf6
None,    # 0xf7
None,    # 0xf8
None,    # 0xf9
None,    # 0xfa
None,    # 0xfb
None,    # 0xfc
None,    # 0xfd
None,    # 0xfe
)



import os

from systools import sys_parsefilepath

from application.repository.pos import CommonPOSRepository
from application.util import LoggerUtil


logger = LoggerUtil.get_logger_name(__name__)


class JanitorService(object):

    def __init__(
        self,                       # type: JanitorService
        common_pos_repository,      # type: CommonPOSRepository
    ):
        # type: (...) -> None

        super(JanitorService, self).__init__()

        self.common_pos_repository = common_pos_repository

        self.data_path = sys_parsefilepath("{HVDATADIR}")
        self.internal_order_path = os.path.join(self.data_path, "bundles", "internalorder")

    def clean_internal_order_file(
        self,       # type: JanitorService
    ):
        # type: (...) -> None

        try:
            if os.path.exists(self.internal_order_path) and os.path.isfile(self.internal_order_path):
                os.remove(self.internal_order_path)
                self.common_pos_repository.restart_hv(
                    event_type="POS_UPDATER_JANITOR",
                )
        except (Exception,):
            logger.exception("Error cleaning internal order file")



# Copyright 2019 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Invocation-side implementation of gRPC Asyncio Python."""

import asyncio
import enum
from functools import partial
import inspect
import logging
import traceback
from typing import Any, AsyncIterator, Generator, Generic, Optional, Tuple

import grpc
from grpc import _common
from grpc._cython import cygrpc

from . import _base_call
from ._metadata import Metadata
from ._typing import DeserializingFunction
from ._typing import DoneCallbackType
from ._typing import MetadatumType
from ._typing import RequestIterableType
from ._typing import RequestType
from ._typing import ResponseType
from ._typing import SerializingFunction

__all__ = "AioRpcError", "Call", "UnaryUnaryCall", "UnaryStreamCall"

_LOCAL_CANCELLATION_DETAILS = "Locally cancelled by application!"
_GC_CANCELLATION_DETAILS = "Cancelled upon garbage collection!"
_RPC_ALREADY_FINISHED_DETAILS = "RPC already finished."
_RPC_HALF_CLOSED_DETAILS = 'RPC is half closed after calling "done_writing".'
_API_STYLE_ERROR = (
    "The iterator and read/write APIs may not be mixed on a single RPC."
)

_OK_CALL_REPRESENTATION = (
    '<{} of RPC that terminated with:\n\tstatus = {}\n\tdetails = "{}"\n>'
)

_NON_OK_CALL_REPRESENTATION = (
    "<{} of RPC that terminated with:\n"
    "\tstatus = {}\n"
    '\tdetails = "{}"\n'
    '\tdebug_error_string = "{}"\n'
    ">"
)

_LOGGER = logging.getLogger(__name__)


class AioRpcError(grpc.RpcError):
    """An implementation of RpcError to be used by the asynchronous API.

    Raised RpcError is a snapshot of the final status of the RPC, values are
    determined. Hence, its methods no longer needs to be coroutines.
    """

    _code: grpc.StatusCode
    _details: Optional[str]
    _initial_metadata: Optional[Metadata]
    _trailing_metadata: Optional[Metadata]
    _debug_error_string: Optional[str]

    def __init__(
        self,
        code: grpc.StatusCode,
        initial_metadata: Metadata,
        trailing_metadata: Metadata,
        details: Optional[str] = None,
        debug_error_string: Optional[str] = None,
    ) -> None:
        """Constructor.

        Args:
          code: The status code with which the RPC has been finalized.
          details: Optional details explaining the reason of the error.
          initial_metadata: Optional initial metadata that could be sent by the
            Server.
          trailing_metadata: Optional metadata that could be sent by the Server.
        """

        super().__init__()
        self._code = code
        self._details = details
        self._initial_metadata = initial_metadata
        self._trailing_metadata = trailing_metadata
        self._debug_error_string = debug_error_string

    def code(self) -> grpc.StatusCode:
        """Accesses the status code sent by the server.

        Returns:
          The `grpc.StatusCode` status code.
        """
        return self._code

    def details(self) -> Optional[str]:
        """Accesses the details sent by the server.

        Returns:
          The description of the error.
        """
        return self._details

    def initial_metadata(self) -> Metadata:
        """Accesses the initial metadata sent by the server.

        Returns:
          The initial metadata received.
        """
        return self._initial_metadata

    def trailing_metadata(self) -> Metadata:
        """Accesses the trailing metadata sent by the server.

        Returns:
          The trailing metadata received.
        """
        return self._trailing_metadata

    def debug_error_string(self) -> str:
        """Accesses the debug error string sent by the server.

        Returns:
          The debug error string received.
        """
        return self._debug_error_string

    def _repr(self) -> str:
        """Assembles the error string for the RPC error."""
        return _NON_OK_CALL_REPRESENTATION.format(
            self.__class__.__name__,
            self._code,
            self._details,
            self._debug_error_string,
        )

    def __repr__(self) -> str:
        return self._repr()

    def __str__(self) -> str:
        return self._repr()

    def __reduce__(self):
        return (
            type(self),
            (
                self._code,
                self._initial_metadata,
                self._trailing_metadata,
                self._details,
                self._debug_error_string,
            ),
        )


def _create_rpc_error(
    initial_metadata: Metadata, status: cygrpc.AioRpcStatus
) -> AioRpcError:
    return AioRpcError(
        _common.CYGRPC_STATUS_CODE_TO_STATUS_CODE[status.code()],
        Metadata.from_tuple(initial_metadata),
        Metadata.from_tuple(status.trailing_metadata()),
        details=status.details(),
        debug_error_string=status.debug_error_string(),
    )


class Call:
    """Base implementation of client RPC Call object.

    Implements logic around final status, metadata and cancellation.
    """

    _loop: asyncio.AbstractEventLoop
    _code: grpc.StatusCode
    _cython_call: cygrpc._AioCall
    _metadata: Tuple[MetadatumType, ...]
    _request_serializer: SerializingFunction
    _response_deserializer: DeserializingFunction

    def __init__(
        self,
        cython_call: cygrpc._AioCall,
        metadata: Metadata,
        request_serializer: SerializingFunction,
        response_deserializer: DeserializingFunction,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        self._loop = loop
        self._cython_call = cython_call
        self._metadata = tuple(metadata)
        self._request_serializer = request_serializer
        self._response_deserializer = response_deserializer

    def __del__(self) -> None:
        # The '_cython_call' object might be destructed before Call object
        if hasattr(self, "_cython_call"):
            if not self._cython_call.done():
                self._cancel(_GC_CANCELLATION_DETAILS)

    def cancelled(self) -> bool:
        return self._cython_call.cancelled()

    def _cancel(self, details: str) -> bool:
        """Forwards the application cancellation reasoning."""
        if not self._cython_call.done():
            self._cython_call.cancel(details)
            return True
        else:
            return False

    def cancel(self) -> bool:
        return self._cancel(_LOCAL_CANCELLATION_DETAILS)

    def done(self) -> bool:
        return self._cython_call.done()

    def add_done_callback(self, callback: DoneCallbackType) -> None:
        cb = partial(callback, self)
        self._cython_call.add_done_callback(cb)

    def time_remaining(self) -> Optional[float]:
        return self._cython_call.time_remaining()

    async def initial_metadata(self) -> Metadata:
        raw_metadata_tuple = await self._cython_call.initial_metadata()
        return Metadata.from_tuple(raw_metadata_tuple)

    async def trailing_metadata(self) -> Metadata:
        raw_metadata_tuple = (
            await self._cython_call.status()
        ).trailing_metadata()
        return Metadata.from_tuple(raw_metadata_tuple)

    async def code(self) -> grpc.StatusCode:
        cygrpc_code = (await self._cython_call.status()).code()
        return _common.CYGRPC_STATUS_CODE_TO_STATUS_CODE[cygrpc_code]

    async def details(self) -> str:
        return (await self._cython_call.status()).details()

    async def debug_error_string(self) -> str:
        return (await self._cython_call.status()).debug_error_string()

    async def _raise_for_status(self) -> None:
        if self._cython_call.is_locally_cancelled():
            raise asyncio.CancelledError()
        code = await self.code()
        if code != grpc.StatusCode.OK:
            raise _create_rpc_error(
                await self.initial_metadata(), await self._cython_call.status()
            )

    def _repr(self) -> str:
        return repr(self._cython_call)

    def __repr__(self) -> str:
        return self._repr()

    def __str__(self) -> str:
        return self._repr()


class _APIStyle(enum.IntEnum):
    UNKNOWN = 0
    ASYNC_GENERATOR = 1
    READER_WRITER = 2


class _UnaryResponseMixin(Call, Generic[ResponseType]):
    _call_response: asyncio.Task

    def _init_unary_response_mixin(self, response_task: asyncio.Task):
        self._call_response = response_task

    def cancel(self) -> bool:
        if super().cancel():
            self._call_response.cancel()
            return True
        else:
            return False

    def __await__(self) -> Generator[Any, None, ResponseType]:
        """Wait till the ongoing RPC request finishes."""
        try:
            response = yield from self._call_response
        except asyncio.CancelledError:
            # Even if we caught all other CancelledError, there is still
            # this corner case. If the application cancels immediately after
            # the Call object is created, we will observe this
            # `CancelledError`.
            if not self.cancelled():
                self.cancel()
            raise

        # NOTE(lidiz) If we raise RpcError in the task, and users doesn't
        # 'await' on it. AsyncIO will log 'Task exception was never retrieved'.
        # Instead, if we move the exception raising here, the spam stops.
        # Unfortunately, there can only be one 'yield from' in '__await__'. So,
        # we need to access the private instance variable.
        if response is cygrpc.EOF:
            if self._cython_call.is_locally_cancelled():
                raise asyncio.CancelledError()
            else:
                raise _create_rpc_error(
                    self._cython_call._initial_metadata,
                    self._cython_call._status,
                )
        else:
            return response


class _StreamResponseMixin(Call):
    _message_aiter: AsyncIterator[ResponseType]
    _preparation: asyncio.Task
    _response_style: _APIStyle

    def _init_stream_response_mixin(self, preparation: asyncio.Task):
        self._message_aiter = None
        self._preparation = preparation
        self._response_style = _APIStyle.UNKNOWN

    def _update_response_style(self, style: _APIStyle):
        if self._response_style is _APIStyle.UNKNOWN:
            self._response_style = style
        elif self._response_style is not style:
            raise cygrpc.UsageError(_API_STYLE_ERROR)

    def cancel(self) -> bool:
        if super().cancel():
            self._preparation.cancel()
            return True
        else:
            return False

    async def _fetch_stream_responses(self) -> ResponseType:
        message = await self._read()
        while message is not cygrpc.EOF:
            yield message
            message = await self._read()

        # If the read operation failed, Core should explain why.
        await self._raise_for_status()

    def __aiter__(self) -> AsyncIterator[ResponseType]:
        self._update_response_style(_APIStyle.ASYNC_GENERATOR)
        if self._message_aiter is None:
            self._message_aiter = self._fetch_stream_responses()
        return self._message_aiter

    async def _read(self) -> ResponseType:
        # Wait for the request being sent
        await self._preparation

        # Reads response message from Core
        try:
            raw_response = await self._cython_call.receive_serialized_message()
        except asyncio.CancelledError:
            if not self.cancelled():
                self.cancel()
            raise

        if raw_response is cygrpc.EOF:
            return cygrpc.EOF
        else:
            return _common.deserialize(
                raw_response, self._response_deserializer
            )

    async def read(self) -> ResponseType:
        if self.done():
            await self._raise_for_status()
            return cygrpc.EOF
        self._update_response_style(_APIStyle.READER_WRITER)

        response_message = await self._read()

        if response_message is cygrpc.EOF:
            # If the read operation failed, Core should explain why.
            await self._raise_for_status()
        return response_message


class _StreamRequestMixin(Call):
    _metadata_sent: asyncio.Event
    _done_writing_flag: bool
    _async_request_poller: Optional[asyncio.Task]
    _request_style: _APIStyle

    def _init_stream_request_mixin(
        self, request_iterator: Optional[RequestIterableType]
    ):
        self._metadata_sent = asyncio.Event()
        self._done_writing_flag = False

        # If user passes in an async iterator, create a consumer Task.
        if request_iterator is not None:
            self._async_request_poller = self._loop.create_task(
                self._consume_request_iterator(request_iterator)
            )
            self._request_style = _APIStyle.ASYNC_GENERATOR
        else:
            self._async_request_poller = None
            self._request_style = _APIStyle.READER_WRITER

    def _raise_for_different_style(self, style: _APIStyle):
        if self._request_style is not style:
            raise cygrpc.UsageError(_API_STYLE_ERROR)

    def cancel(self) -> bool:
        if super().cancel():
            if self._async_request_poller is not None:
                self._async_request_poller.cancel()
            return True
        else:
            return False

    def _metadata_sent_observer(self):
        self._metadata_sent.set()

    async def _consume_request_iterator(
        self, request_iterator: RequestIterableType
    ) -> None:
        try:
            if inspect.isasyncgen(request_iterator) or hasattr(
                request_iterator, "__aiter__"
            ):
                async for request in request_iterator:
                    try:
                        await self._write(request)
                    except AioRpcError as rpc_error:
                        _LOGGER.debug(
                            (
                                "Exception while consuming the"
                                " request_iterator: %s"
                            ),
                            rpc_error,
                        )
                        return
            else:
                for request in request_iterator:
                    try:
                        await self._write(request)
                    except AioRpcError as rpc_error:
                        _LOGGER.debug(
                            (
                                "Exception while consuming the"
                                " request_iterator: %s"
                            ),
                            rpc_error,
                        )
                        return

            await self._done_writing()
        except:  # pylint: disable=bare-except
            # Client iterators can raise exceptions, which we should handle by
            # cancelling the RPC and logging the client's error. No exceptions
            # should escape this function.
            _LOGGER.debug(
                "Client request_iterator raised exception:\n%s",
                traceback.format_exc(),
            )
            self.cancel()

    async def _write(self, request: RequestType) -> None:
        if self.done():
            raise asyncio.InvalidStateError(_RPC_ALREADY_FINISHED_DETAILS)
        if self._done_writing_flag:
            raise asyncio.InvalidStateError(_RPC_HALF_CLOSED_DETAILS)
        if not self._metadata_sent.is_set():
            await self._metadata_sent.wait()
            if self.done():
                await self._raise_for_status()

        serialized_request = _common.serialize(
            request, self._request_serializer
        )
        try:
            await self._cython_call.send_serialized_message(serialized_request)
        except cygrpc.InternalError as err:
            self._cython_call.set_internal_error(str(err))
            await self._raise_for_status()
        except asyncio.CancelledError:
            if not self.cancelled():
                self.cancel()
            raise

    async def _done_writing(self) -> None:
        if self.done():
            # If the RPC is finished, do nothing.
            return
        if not self._done_writing_flag:
            # If the done writing is not sent before, try to send it.
            self._done_writing_flag = True
            try:
                await self._cython_call.send_receive_close()
            except asyncio.CancelledError:
                if not self.cancelled():
                    self.cancel()
                raise

    async def write(self, request: RequestType) -> None:
        self._raise_for_different_style(_APIStyle.READER_WRITER)
        await self._write(request)

    async def done_writing(self) -> None:
        """Signal peer that client is done writing.

        This method is idempotent.
        """
        self._raise_for_different_style(_APIStyle.READER_WRITER)
        await self._done_writing()

    async def wait_for_connection(self) -> None:
        await self._metadata_sent.wait()
        if self.done():
            await self._raise_for_status()


class UnaryUnaryCall(_UnaryResponseMixin, Call, _base_call.UnaryUnaryCall):
    """Object for managing unary-unary RPC calls.

    Returned when an instance of `UnaryUnaryMultiCallable` object is called.
    """

    _request: RequestType
    _invocation_task: asyncio.Task

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        request: RequestType,
        deadline: Optional[float],
        metadata: Metadata,
        credentials: Optional[grpc.CallCredentials],
        wait_for_ready: Optional[bool],
        channel: cygrpc.AioChannel,
        method: bytes,
        request_serializer: SerializingFunction,
        response_deserializer: DeserializingFunction,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        super().__init__(
            channel.call(method, deadline, credentials, wait_for_ready),
            metadata,
            request_serializer,
            response_deserializer,
            loop,
        )
        self._request = request
        self._context = cygrpc.build_census_context()
        self._invocation_task = loop.create_task(self._invoke())
        self._init_unary_response_mixin(self._invocation_task)

    async def _invoke(self) -> ResponseType:
        serialized_request = _common.serialize(
            self._request, self._request_serializer
        )

        # NOTE(lidiz) asyncio.CancelledError is not a good transport for status,
        # because the asyncio.Task class do not cache the exception object.
        # https://github.com/python/cpython/blob/edad4d89e357c92f70c0324b937845d652b20afd/Lib/asyncio/tasks.py#L785
        try:
            serialized_response = await self._cython_call.unary_unary(
                serialized_request, self._metadata, self._context
            )
        except asyncio.CancelledError:
            if not self.cancelled():
                self.cancel()

        if self._cython_call.is_ok():
            return _common.deserialize(
                serialized_response, self._response_deserializer
            )
        else:
            return cygrpc.EOF

    async def wait_for_connection(self) -> None:
        await self._invocation_task
        if self.done():
            await self._raise_for_status()


class UnaryStreamCall(_StreamResponseMixin, Call, _base_call.UnaryStreamCall):
    """Object for managing unary-stream RPC calls.

    Returned when an instance of `UnaryStreamMultiCallable` object is called.
    """

    _request: RequestType
    _send_unary_request_task: asyncio.Task

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        request: RequestType,
        deadline: Optional[float],
        metadata: Metadata,
        credentials: Optional[grpc.CallCredentials],
        wait_for_ready: Optional[bool],
        channel: cygrpc.AioChannel,
        method: bytes,
        request_serializer: SerializingFunction,
        response_deserializer: DeserializingFunction,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        super().__init__(
            channel.call(method, deadline, credentials, wait_for_ready),
            metadata,
            request_serializer,
            response_deserializer,
            loop,
        )
        self._request = request
        self._context = cygrpc.build_census_context()
        self._send_unary_request_task = loop.create_task(
            self._send_unary_request()
        )
        self._init_stream_response_mixin(self._send_unary_request_task)

    async def _send_unary_request(self) -> ResponseType:
        serialized_request = _common.serialize(
            self._request, self._request_serializer
        )
        try:
            await self._cython_call.initiate_unary_stream(
                serialized_request, self._metadata, self._context
            )
        except asyncio.CancelledError:
            if not self.cancelled():
                self.cancel()
            raise

    async def wait_for_connection(self) -> None:
        await self._send_unary_request_task
        if self.done():
            await self._raise_for_status()


# pylint: disable=too-many-ancestors
class StreamUnaryCall(
    _StreamRequestMixin, _UnaryResponseMixin, Call, _base_call.StreamUnaryCall
):
    """Object for managing stream-unary RPC calls.

    Returned when an instance of `StreamUnaryMultiCallable` object is called.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        request_iterator: Optional[RequestIterableType],
        deadline: Optional[float],
        metadata: Metadata,
        credentials: Optional[grpc.CallCredentials],
        wait_for_ready: Optional[bool],
        channel: cygrpc.AioChannel,
        method: bytes,
        request_serializer: SerializingFunction,
        response_deserializer: DeserializingFunction,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        super().__init__(
            channel.call(method, deadline, credentials, wait_for_ready),
            metadata,
            request_serializer,
            response_deserializer,
            loop,
        )

        self._context = cygrpc.build_census_context()
        self._init_stream_request_mixin(request_iterator)
        self._init_unary_response_mixin(loop.create_task(self._conduct_rpc()))

    async def _conduct_rpc(self) -> ResponseType:
        try:
            serialized_response = await self._cython_call.stream_unary(
                self._metadata, self._metadata_sent_observer, self._context
            )
        except asyncio.CancelledError:
            if not self.cancelled():
                self.cancel()
            raise

        if self._cython_call.is_ok():
            return _common.deserialize(
                serialized_response, self._response_deserializer
            )
        else:
            return cygrpc.EOF


class StreamStreamCall(
    _StreamRequestMixin, _StreamResponseMixin, Call, _base_call.StreamStreamCall
):
    """Object for managing stream-stream RPC calls.

    Returned when an instance of `StreamStreamMultiCallable` object is called.
    """

    _initializer: asyncio.Task

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        request_iterator: Optional[RequestIterableType],
        deadline: Optional[float],
        metadata: Metadata,
        credentials: Optional[grpc.CallCredentials],
        wait_for_ready: Optional[bool],
        channel: cygrpc.AioChannel,
        method: bytes,
        request_serializer: SerializingFunction,
        response_deserializer: DeserializingFunction,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        super().__init__(
            channel.call(method, deadline, credentials, wait_for_ready),
            metadata,
            request_serializer,
            response_deserializer,
            loop,
        )
        self._context = cygrpc.build_census_context()
        self._initializer = self._loop.create_task(self._prepare_rpc())
        self._init_stream_request_mixin(request_iterator)
        self._init_stream_response_mixin(self._initializer)

    async def _prepare_rpc(self):
        """This method prepares the RPC for receiving/sending messages.

        All other operations around the stream should only happen after the
        completion of this method.
        """
        try:
            await self._cython_call.initiate_stream_stream(
                self._metadata, self._metadata_sent_observer, self._context
            )
        except asyncio.CancelledError:
            if not self.cancelled():
                self.cancel()
            # No need to raise RpcError here, because no one will `await` this task.



data = (
'ddyels',    # 0x00
'ddyelt',    # 0x01
'ddyelp',    # 0x02
'ddyelh',    # 0x03
'ddyem',    # 0x04
'ddyeb',    # 0x05
'ddyebs',    # 0x06
'ddyes',    # 0x07
'ddyess',    # 0x08
'ddyeng',    # 0x09
'ddyej',    # 0x0a
'ddyec',    # 0x0b
'ddyek',    # 0x0c
'ddyet',    # 0x0d
'ddyep',    # 0x0e
'ddyeh',    # 0x0f
'ddo',    # 0x10
'ddog',    # 0x11
'ddogg',    # 0x12
'ddogs',    # 0x13
'ddon',    # 0x14
'ddonj',    # 0x15
'ddonh',    # 0x16
'ddod',    # 0x17
'ddol',    # 0x18
'ddolg',    # 0x19
'ddolm',    # 0x1a
'ddolb',    # 0x1b
'ddols',    # 0x1c
'ddolt',    # 0x1d
'ddolp',    # 0x1e
'ddolh',    # 0x1f
'ddom',    # 0x20
'ddob',    # 0x21
'ddobs',    # 0x22
'ddos',    # 0x23
'ddoss',    # 0x24
'ddong',    # 0x25
'ddoj',    # 0x26
'ddoc',    # 0x27
'ddok',    # 0x28
'ddot',    # 0x29
'ddop',    # 0x2a
'ddoh',    # 0x2b
'ddwa',    # 0x2c
'ddwag',    # 0x2d
'ddwagg',    # 0x2e
'ddwags',    # 0x2f
'ddwan',    # 0x30
'ddwanj',    # 0x31
'ddwanh',    # 0x32
'ddwad',    # 0x33
'ddwal',    # 0x34
'ddwalg',    # 0x35
'ddwalm',    # 0x36
'ddwalb',    # 0x37
'ddwals',    # 0x38
'ddwalt',    # 0x39
'ddwalp',    # 0x3a
'ddwalh',    # 0x3b
'ddwam',    # 0x3c
'ddwab',    # 0x3d
'ddwabs',    # 0x3e
'ddwas',    # 0x3f
'ddwass',    # 0x40
'ddwang',    # 0x41
'ddwaj',    # 0x42
'ddwac',    # 0x43
'ddwak',    # 0x44
'ddwat',    # 0x45
'ddwap',    # 0x46
'ddwah',    # 0x47
'ddwae',    # 0x48
'ddwaeg',    # 0x49
'ddwaegg',    # 0x4a
'ddwaegs',    # 0x4b
'ddwaen',    # 0x4c
'ddwaenj',    # 0x4d
'ddwaenh',    # 0x4e
'ddwaed',    # 0x4f
'ddwael',    # 0x50
'ddwaelg',    # 0x51
'ddwaelm',    # 0x52
'ddwaelb',    # 0x53
'ddwaels',    # 0x54
'ddwaelt',    # 0x55
'ddwaelp',    # 0x56
'ddwaelh',    # 0x57
'ddwaem',    # 0x58
'ddwaeb',    # 0x59
'ddwaebs',    # 0x5a
'ddwaes',    # 0x5b
'ddwaess',    # 0x5c
'ddwaeng',    # 0x5d
'ddwaej',    # 0x5e
'ddwaec',    # 0x5f
'ddwaek',    # 0x60
'ddwaet',    # 0x61
'ddwaep',    # 0x62
'ddwaeh',    # 0x63
'ddoe',    # 0x64
'ddoeg',    # 0x65
'ddoegg',    # 0x66
'ddoegs',    # 0x67
'ddoen',    # 0x68
'ddoenj',    # 0x69
'ddoenh',    # 0x6a
'ddoed',    # 0x6b
'ddoel',    # 0x6c
'ddoelg',    # 0x6d
'ddoelm',    # 0x6e
'ddoelb',    # 0x6f
'ddoels',    # 0x70
'ddoelt',    # 0x71
'ddoelp',    # 0x72
'ddoelh',    # 0x73
'ddoem',    # 0x74
'ddoeb',    # 0x75
'ddoebs',    # 0x76
'ddoes',    # 0x77
'ddoess',    # 0x78
'ddoeng',    # 0x79
'ddoej',    # 0x7a
'ddoec',    # 0x7b
'ddoek',    # 0x7c
'ddoet',    # 0x7d
'ddoep',    # 0x7e
'ddoeh',    # 0x7f
'ddyo',    # 0x80
'ddyog',    # 0x81
'ddyogg',    # 0x82
'ddyogs',    # 0x83
'ddyon',    # 0x84
'ddyonj',    # 0x85
'ddyonh',    # 0x86
'ddyod',    # 0x87
'ddyol',    # 0x88
'ddyolg',    # 0x89
'ddyolm',    # 0x8a
'ddyolb',    # 0x8b
'ddyols',    # 0x8c
'ddyolt',    # 0x8d
'ddyolp',    # 0x8e
'ddyolh',    # 0x8f
'ddyom',    # 0x90
'ddyob',    # 0x91
'ddyobs',    # 0x92
'ddyos',    # 0x93
'ddyoss',    # 0x94
'ddyong',    # 0x95
'ddyoj',    # 0x96
'ddyoc',    # 0x97
'ddyok',    # 0x98
'ddyot',    # 0x99
'ddyop',    # 0x9a
'ddyoh',    # 0x9b
'ddu',    # 0x9c
'ddug',    # 0x9d
'ddugg',    # 0x9e
'ddugs',    # 0x9f
'ddun',    # 0xa0
'ddunj',    # 0xa1
'ddunh',    # 0xa2
'ddud',    # 0xa3
'ddul',    # 0xa4
'ddulg',    # 0xa5
'ddulm',    # 0xa6
'ddulb',    # 0xa7
'dduls',    # 0xa8
'ddult',    # 0xa9
'ddulp',    # 0xaa
'ddulh',    # 0xab
'ddum',    # 0xac
'ddub',    # 0xad
'ddubs',    # 0xae
'ddus',    # 0xaf
'dduss',    # 0xb0
'ddung',    # 0xb1
'dduj',    # 0xb2
'dduc',    # 0xb3
'dduk',    # 0xb4
'ddut',    # 0xb5
'ddup',    # 0xb6
'dduh',    # 0xb7
'ddweo',    # 0xb8
'ddweog',    # 0xb9
'ddweogg',    # 0xba
'ddweogs',    # 0xbb
'ddweon',    # 0xbc
'ddweonj',    # 0xbd
'ddweonh',    # 0xbe
'ddweod',    # 0xbf
'ddweol',    # 0xc0
'ddweolg',    # 0xc1
'ddweolm',    # 0xc2
'ddweolb',    # 0xc3
'ddweols',    # 0xc4
'ddweolt',    # 0xc5
'ddweolp',    # 0xc6
'ddweolh',    # 0xc7
'ddweom',    # 0xc8
'ddweob',    # 0xc9
'ddweobs',    # 0xca
'ddweos',    # 0xcb
'ddweoss',    # 0xcc
'ddweong',    # 0xcd
'ddweoj',    # 0xce
'ddweoc',    # 0xcf
'ddweok',    # 0xd0
'ddweot',    # 0xd1
'ddweop',    # 0xd2
'ddweoh',    # 0xd3
'ddwe',    # 0xd4
'ddweg',    # 0xd5
'ddwegg',    # 0xd6
'ddwegs',    # 0xd7
'ddwen',    # 0xd8
'ddwenj',    # 0xd9
'ddwenh',    # 0xda
'ddwed',    # 0xdb
'ddwel',    # 0xdc
'ddwelg',    # 0xdd
'ddwelm',    # 0xde
'ddwelb',    # 0xdf
'ddwels',    # 0xe0
'ddwelt',    # 0xe1
'ddwelp',    # 0xe2
'ddwelh',    # 0xe3
'ddwem',    # 0xe4
'ddweb',    # 0xe5
'ddwebs',    # 0xe6
'ddwes',    # 0xe7
'ddwess',    # 0xe8
'ddweng',    # 0xe9
'ddwej',    # 0xea
'ddwec',    # 0xeb
'ddwek',    # 0xec
'ddwet',    # 0xed
'ddwep',    # 0xee
'ddweh',    # 0xef
'ddwi',    # 0xf0
'ddwig',    # 0xf1
'ddwigg',    # 0xf2
'ddwigs',    # 0xf3
'ddwin',    # 0xf4
'ddwinj',    # 0xf5
'ddwinh',    # 0xf6
'ddwid',    # 0xf7
'ddwil',    # 0xf8
'ddwilg',    # 0xf9
'ddwilm',    # 0xfa
'ddwilb',    # 0xfb
'ddwils',    # 0xfc
'ddwilt',    # 0xfd
'ddwilp',    # 0xfe
'ddwilh',    # 0xff
)



# siren.py - functions for handling French SIREN numbers
# coding: utf-8
#
# Copyright (C) 2012-2017 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""SIREN (a French company identification number).

The SIREN (Système d'Identification du Répertoire des Entreprises) is a 9
digit number used to identify French companies. The Luhn checksum is used
to validate the numbers.

>>> compact('552 008 443')
'552008443'
>>> validate('404833048')
'404833048'
>>> validate('404833047')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> to_tva('443 121 975')
'46 443 121 975'
"""

from stdnum import luhn
from stdnum.exceptions import *
from stdnum.util import clean, isdigits


# An online validation function is available but it does not provide an
# automated entry point, has usage restrictions and seems to require
# attribution to the service for any results used.
# https://avis-situation-sirene.insee.fr/


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' .').strip()


def validate(number):
    """Check if the number provided is a valid SIREN. This checks the length,
    formatting and check digit."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if len(number) != 9:
        raise InvalidLength()
    luhn.validate(number)
    return number


def is_valid(number):
    """Check if the number provided is a valid SIREN."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def to_tva(number):
    """Return a TVA that prepends the two extra check digits to the SIREN."""
    # note that this always returns numeric check digits
    # it is unclean when the alphabetic ones are used
    return '%02d%s%s' % (
        int(compact(number) + '12') % 97,
        ' ' if ' ' in number else '',
        number)



class MessageBusError(Exception):
    def __init__(
        self,   # type: MessageBusError
        component,  # type: str
        token,      # type: int
        data,       # type: bytes
    ):
        self.component = component
        self.token = token
        self.data = data

    def __str__(self):
        return "Error sending message to {}. Response token: {}, response data: {}".format(
            self.component,
            self.token,
            self.data
        )

    def __repr__(self):
        return self.__str__()



import json

from messagebus import NakMessage, AckMessage
from messageprocessorutil import UuidMessageProcessor
from production.manager import ProductionManager


class StatisticUpdateProcessor(UuidMessageProcessor):
    def __init__(self, production_manager):
        # type: (ProductionManager) -> None
        super(StatisticUpdateProcessor, self).__init__()
        self.production_manager = production_manager

    def parse_data(self, data, event_type):
        # This processor receives no data
        pass

    def call_business(self, _):
        return self.production_manager.statistics or {}

    def format_response(self, message_bus, message, event, input_obj, result):
        message_bus.reply_message(message, AckMessage(json.dumps(result)))

    def format_exception(self, message_bus, message, event, input_obj, exception):
        message_bus.reply_message(message, NakMessage(exception.message))

    def format_parse_exception(self, message_bus, message, event, data, exception):
        message_bus.reply_message(message, NakMessage(exception.message))



# pps.py - functions for handling Irish PPS numbers
#
# Copyright (C) 2012, 2013 Arthur de Jong
# Copyright (C) 2014 Olivier Dony
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""PPS No (Personal Public Service Number, Irish personal number).

The Personal Public Service number consists of 7 digits, and one or
two letters. The first letter is a check character.
When present (which should be the case for new numbers as of 2013),
the second letter can be 'A' (for individuals) or 'H' (for
non-individuals, such as limited companies, trusts, partnerships
and unincorporated bodies). Pre-2013 values may have 'W', 'T',
or 'X' as the second letter ; it is ignored by the check.

>>> validate('6433435F')  # pre-2013
'6433435F'
>>> validate('6433435FT')  # pre-2013 with special final 'T'
'6433435FT'
>>> validate('6433435FW')  # pre-2013 format for married women
'6433435FW'
>>> validate('6433435E')  # incorrect check digit
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('6433435OA')  # 2013 format (personal)
'6433435OA'
>>> validate('6433435IH')  # 2013 format (non-personal)
'6433435IH'
>>> validate('6433435VH')  # 2013 format (incorrect check)
Traceback (most recent call last):
    ...
InvalidChecksum: ...
"""

import re

from stdnum.exceptions import *
from stdnum.ie import vat
from stdnum.util import clean


pps_re = re.compile(r'^\d{7}[A-W][AHWTX]?$')
"""Regular expression used to check syntax of PPS numbers."""


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -').upper().strip()


def validate(number):
    """Check if the number provided is a valid PPS number. This checks the
    length, formatting and check digit."""
    number = compact(number)
    if not pps_re.match(number):
        raise InvalidFormat()
    if len(number) == 9 and number[8] in 'AH':
        # new 2013 format
        if number[7] != vat.calc_check_digit(number[:7] + number[8:]):
            raise InvalidChecksum()
    else:
        # old format, last letter ignored
        if number[7] != vat.calc_check_digit(number[:7]):
            raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number provided is a valid PPS number. This checks the
    length, formatting and check digit."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



# -*- coding: utf-8 -*-
"""
This module provides an interface to the native time zone data on Windows,
including :py:class:`datetime.tzinfo` implementations.

Attempting to import this module on a non-Windows platform will raise an
:py:obj:`ImportError`.
"""
# This code was originally contributed by Jeffrey Harris.
import datetime
import struct

from six.moves import winreg
from six import text_type

try:
    import ctypes
    from ctypes import wintypes
except ValueError:
    # ValueError is raised on non-Windows systems for some horrible reason.
    raise ImportError("Running tzwin on non-Windows system")

from ._common import tzrangebase

__all__ = ["tzwin", "tzwinlocal", "tzres"]

ONEWEEK = datetime.timedelta(7)

TZKEYNAMENT = r"SOFTWARE\Microsoft\Windows NT\CurrentVersion\Time Zones"
TZKEYNAME9X = r"SOFTWARE\Microsoft\Windows\CurrentVersion\Time Zones"
TZLOCALKEYNAME = r"SYSTEM\CurrentControlSet\Control\TimeZoneInformation"


def _settzkeyname():
    handle = winreg.ConnectRegistry(None, winreg.HKEY_LOCAL_MACHINE)
    try:
        winreg.OpenKey(handle, TZKEYNAMENT).Close()
        TZKEYNAME = TZKEYNAMENT
    except WindowsError:
        TZKEYNAME = TZKEYNAME9X
    handle.Close()
    return TZKEYNAME


TZKEYNAME = _settzkeyname()


class tzres(object):
    """
    Class for accessing ``tzres.dll``, which contains timezone name related
    resources.

    .. versionadded:: 2.5.0
    """
    p_wchar = ctypes.POINTER(wintypes.WCHAR)        # Pointer to a wide char

    def __init__(self, tzres_loc='tzres.dll'):
        # Load the user32 DLL so we can load strings from tzres
        user32 = ctypes.WinDLL('user32')

        # Specify the LoadStringW function
        user32.LoadStringW.argtypes = (wintypes.HINSTANCE,
                                       wintypes.UINT,
                                       wintypes.LPWSTR,
                                       ctypes.c_int)

        self.LoadStringW = user32.LoadStringW
        self._tzres = ctypes.WinDLL(tzres_loc)
        self.tzres_loc = tzres_loc

    def load_name(self, offset):
        """
        Load a timezone name from a DLL offset (integer).

        >>> from dateutil.tzwin import tzres
        >>> tzr = tzres()
        >>> print(tzr.load_name(112))
        'Eastern Standard Time'

        :param offset:
            A positive integer value referring to a string from the tzres dll.

        .. note::

            Offsets found in the registry are generally of the form
            ``@tzres.dll,-114``. The offset in this case is 114, not -114.

        """
        resource = self.p_wchar()
        lpBuffer = ctypes.cast(ctypes.byref(resource), wintypes.LPWSTR)
        nchar = self.LoadStringW(self._tzres._handle, offset, lpBuffer, 0)
        return resource[:nchar]

    def name_from_string(self, tzname_str):
        """
        Parse strings as returned from the Windows registry into the time zone
        name as defined in the registry.

        >>> from dateutil.tzwin import tzres
        >>> tzr = tzres()
        >>> print(tzr.name_from_string('@tzres.dll,-251'))
        'Dateline Daylight Time'
        >>> print(tzr.name_from_string('Eastern Standard Time'))
        'Eastern Standard Time'

        :param tzname_str:
            A timezone name string as returned from a Windows registry key.

        :return:
            Returns the localized timezone string from tzres.dll if the string
            is of the form `@tzres.dll,-offset`, else returns the input string.
        """
        if not tzname_str.startswith('@'):
            return tzname_str

        name_splt = tzname_str.split(',-')
        try:
            offset = int(name_splt[1])
        except:
            raise ValueError("Malformed timezone string.")

        return self.load_name(offset)


class tzwinbase(tzrangebase):
    """tzinfo class based on win32's timezones available in the registry."""
    def __init__(self):
        raise NotImplementedError('tzwinbase is an abstract base class')

    def __eq__(self, other):
        # Compare on all relevant dimensions, including name.
        if not isinstance(other, tzwinbase):
            return NotImplemented

        return  (self._std_offset == other._std_offset and
                 self._dst_offset == other._dst_offset and
                 self._stddayofweek == other._stddayofweek and
                 self._dstdayofweek == other._dstdayofweek and
                 self._stdweeknumber == other._stdweeknumber and
                 self._dstweeknumber == other._dstweeknumber and
                 self._stdhour == other._stdhour and
                 self._dsthour == other._dsthour and
                 self._stdminute == other._stdminute and
                 self._dstminute == other._dstminute and
                 self._std_abbr == other._std_abbr and
                 self._dst_abbr == other._dst_abbr)

    @staticmethod
    def list():
        """Return a list of all time zones known to the system."""
        with winreg.ConnectRegistry(None, winreg.HKEY_LOCAL_MACHINE) as handle:
            with winreg.OpenKey(handle, TZKEYNAME) as tzkey:
                result = [winreg.EnumKey(tzkey, i)
                          for i in range(winreg.QueryInfoKey(tzkey)[0])]
        return result

    def display(self):
        """
        Return the display name of the time zone.
        """
        return self._display

    def transitions(self, year):
        """
        For a given year, get the DST on and off transition times, expressed
        always on the standard time side. For zones with no transitions, this
        function returns ``None``.

        :param year:
            The year whose transitions you would like to query.

        :return:
            Returns a :class:`tuple` of :class:`datetime.datetime` objects,
            ``(dston, dstoff)`` for zones with an annual DST transition, or
            ``None`` for fixed offset zones.
        """

        if not self.hasdst:
            return None

        dston = picknthweekday(year, self._dstmonth, self._dstdayofweek,
                               self._dsthour, self._dstminute,
                               self._dstweeknumber)

        dstoff = picknthweekday(year, self._stdmonth, self._stddayofweek,
                                self._stdhour, self._stdminute,
                                self._stdweeknumber)

        # Ambiguous dates default to the STD side
        dstoff -= self._dst_base_offset

        return dston, dstoff

    def _get_hasdst(self):
        return self._dstmonth != 0

    @property
    def _dst_base_offset(self):
        return self._dst_base_offset_


class tzwin(tzwinbase):
    """
    Time zone object created from the zone info in the Windows registry

    These are similar to :py:class:`dateutil.tz.tzrange` objects in that
    the time zone data is provided in the format of a single offset rule
    for either 0 or 2 time zone transitions per year.

    :param: name
        The name of a Windows time zone key, e.g. "Eastern Standard Time".
        The full list of keys can be retrieved with :func:`tzwin.list`.
    """

    def __init__(self, name):
        self._name = name

        with winreg.ConnectRegistry(None, winreg.HKEY_LOCAL_MACHINE) as handle:
            tzkeyname = text_type("{kn}\\{name}").format(kn=TZKEYNAME, name=name)
            with winreg.OpenKey(handle, tzkeyname) as tzkey:
                keydict = valuestodict(tzkey)

        self._std_abbr = keydict["Std"]
        self._dst_abbr = keydict["Dlt"]

        self._display = keydict["Display"]

        # See http://ww_winreg.jsiinc.com/SUBA/tip0300/rh0398.htm
        tup = struct.unpack("=3l16h", keydict["TZI"])
        stdoffset = -tup[0]-tup[1]          # Bias + StandardBias * -1
        dstoffset = stdoffset-tup[2]        # + DaylightBias * -1
        self._std_offset = datetime.timedelta(minutes=stdoffset)
        self._dst_offset = datetime.timedelta(minutes=dstoffset)

        # for the meaning see the win32 TIME_ZONE_INFORMATION structure docs
        # http://msdn.microsoft.com/en-us/library/windows/desktop/ms725481(v=vs.85).aspx
        (self._stdmonth,
         self._stddayofweek,   # Sunday = 0
         self._stdweeknumber,  # Last = 5
         self._stdhour,
         self._stdminute) = tup[4:9]

        (self._dstmonth,
         self._dstdayofweek,   # Sunday = 0
         self._dstweeknumber,  # Last = 5
         self._dsthour,
         self._dstminute) = tup[12:17]

        self._dst_base_offset_ = self._dst_offset - self._std_offset
        self.hasdst = self._get_hasdst()

    def __repr__(self):
        return "tzwin(%s)" % repr(self._name)

    def __reduce__(self):
        return (self.__class__, (self._name,))


class tzwinlocal(tzwinbase):
    """
    Class representing the local time zone information in the Windows registry

    While :class:`dateutil.tz.tzlocal` makes system calls (via the :mod:`time`
    module) to retrieve time zone information, ``tzwinlocal`` retrieves the
    rules directly from the Windows registry and creates an object like
    :class:`dateutil.tz.tzwin`.

    Because Windows does not have an equivalent of :func:`time.tzset`, on
    Windows, :class:`dateutil.tz.tzlocal` instances will always reflect the
    time zone settings *at the time that the process was started*, meaning
    changes to the machine's time zone settings during the run of a program
    on Windows will **not** be reflected by :class:`dateutil.tz.tzlocal`.
    Because ``tzwinlocal`` reads the registry directly, it is unaffected by
    this issue.
    """
    def __init__(self):
        with winreg.ConnectRegistry(None, winreg.HKEY_LOCAL_MACHINE) as handle:
            with winreg.OpenKey(handle, TZLOCALKEYNAME) as tzlocalkey:
                keydict = valuestodict(tzlocalkey)

            self._std_abbr = keydict["StandardName"]
            self._dst_abbr = keydict["DaylightName"]

            try:
                tzkeyname = text_type('{kn}\\{sn}').format(kn=TZKEYNAME,
                                                          sn=self._std_abbr)
                with winreg.OpenKey(handle, tzkeyname) as tzkey:
                    _keydict = valuestodict(tzkey)
                    self._display = _keydict["Display"]
            except OSError:
                self._display = None

        stdoffset = -keydict["Bias"]-keydict["StandardBias"]
        dstoffset = stdoffset-keydict["DaylightBias"]

        self._std_offset = datetime.timedelta(minutes=stdoffset)
        self._dst_offset = datetime.timedelta(minutes=dstoffset)

        # For reasons unclear, in this particular key, the day of week has been
        # moved to the END of the SYSTEMTIME structure.
        tup = struct.unpack("=8h", keydict["StandardStart"])

        (self._stdmonth,
         self._stdweeknumber,  # Last = 5
         self._stdhour,
         self._stdminute) = tup[1:5]

        self._stddayofweek = tup[7]

        tup = struct.unpack("=8h", keydict["DaylightStart"])

        (self._dstmonth,
         self._dstweeknumber,  # Last = 5
         self._dsthour,
         self._dstminute) = tup[1:5]

        self._dstdayofweek = tup[7]

        self._dst_base_offset_ = self._dst_offset - self._std_offset
        self.hasdst = self._get_hasdst()

    def __repr__(self):
        return "tzwinlocal()"

    def __str__(self):
        # str will return the standard name, not the daylight name.
        return "tzwinlocal(%s)" % repr(self._std_abbr)

    def __reduce__(self):
        return (self.__class__, ())


def picknthweekday(year, month, dayofweek, hour, minute, whichweek):
    """ dayofweek == 0 means Sunday, whichweek 5 means last instance """
    first = datetime.datetime(year, month, 1, hour, minute)

    # This will work if dayofweek is ISO weekday (1-7) or Microsoft-style (0-6),
    # Because 7 % 7 = 0
    weekdayone = first.replace(day=((dayofweek - first.isoweekday()) % 7) + 1)
    wd = weekdayone + ((whichweek - 1) * ONEWEEK)
    if (wd.month != month):
        wd -= ONEWEEK

    return wd


def valuestodict(key):
    """Convert a registry key's values to a dictionary."""
    dout = {}
    size = winreg.QueryInfoKey(key)[1]
    tz_res = None

    for i in range(size):
        key_name, value, dtype = winreg.EnumValue(key, i)
        if dtype == winreg.REG_DWORD or dtype == winreg.REG_DWORD_LITTLE_ENDIAN:
            # If it's a DWORD (32-bit integer), it's stored as unsigned - convert
            # that to a proper signed integer
            if value & (1 << 31):
                value = value - (1 << 32)
        elif dtype == winreg.REG_SZ:
            # If it's a reference to the tzres DLL, load the actual string
            if value.startswith('@tzres'):
                tz_res = tz_res or tzres()
                value = tz_res.name_from_string(value)

            value = value.rstrip('\x00')    # Remove trailing nulls

        dout[key_name] = value

    return dout



from typing import (
    Any,
    Dict,
)

from nreports.builders import DataBuilder
from nreports_api.shared import JSON


class BaseLoyaltyCoupon(DataBuilder):

    def build_data(
        self,
        data_to_build: Dict[str, JSON],
        **kwargs: Any,
    ) -> Dict[str, JSON]:

        raise NotImplementedError()

    def fill_constant_values(
        self,
        pos_id: int,
        order_id: int,
        data_to_build: Dict[str, JSON],
    ) -> None:

        data_to_build["LOYALTY_TEXT"] = self.constants.get("LOYALTY_TEXT", "")
        data_to_build["PRINT_LOYALTY_COUPON"] = data_to_build["LOYALTY_TEXT"] != ""

        loyalty_coupon_before_danfe = self.constants.get("LOYALTY_COUPON_DANFE_POSITION", "before").lower() == "before"
        data_to_build["LOYALTY_COUPON_BEFORE_DANFE"] = loyalty_coupon_before_danfe
        data_to_build["LOYALTY_COUPON_AFTER_DANFE"] = not loyalty_coupon_before_danfe

        order_service = self.order_api_repository.get_order_service(pos_id=pos_id)
        order_xml = order_service.get_order_picture(order_id=order_id)

        is_delivery_order = order_xml.get("saleType", "") == "3"
        allow_delivery_orders = self.constants.get("LOYALTY_COUPON_FOR_DELIVERY_ORDER", "false").lower() == "true"
        if is_delivery_order and allow_delivery_orders is False:
            data_to_build["LOYALTY_COUPON_BEFORE_DANFE"] = False
            data_to_build["LOYALTY_COUPON_AFTER_DANFE"] = False



import os
import re
import sys
import warnings
from datetime import timezone
import pytz_deprecation_shim as pds

from tzlocal import utils

if sys.version_info >= (3, 9):
    from zoneinfo import ZoneInfo  # pragma: no cover
else:
    from backports.zoneinfo import ZoneInfo  # pragma: no cover

_cache_tz = None
_cache_tz_name = None


def _get_localzone_name(_root="/"):
    """Tries to find the local timezone configuration.

    This method finds the timezone name, if it can, or it returns None.

    The parameter _root makes the function look for files like /etc/localtime
    beneath the _root directory. This is primarily used by the tests.
    In normal usage you call the function without parameters."""

    # First try the ENV setting.
    tzenv = utils._tz_name_from_env()
    if tzenv:
        return tzenv

    # Are we under Termux on Android?
    if os.path.exists(os.path.join(_root, "system/bin/getprop")):
        import subprocess

        androidtz = (
            subprocess.check_output(["getprop", "persist.sys.timezone"])
            .strip()
            .decode()
        )
        return androidtz

    # Now look for distribution specific configuration files
    # that contain the timezone name.

    # Stick all of them in a dict, to compare later.
    found_configs = {}

    for configfile in ("etc/timezone", "var/db/zoneinfo"):
        tzpath = os.path.join(_root, configfile)
        try:
            with open(tzpath, "rt") as tzfile:
                data = tzfile.read()

                etctz = data.strip('/ \t\r\n')
                if not etctz:
                    # Empty file, skip
                    continue
                for etctz in etctz.splitlines():
                    # Get rid of host definitions and comments:
                    if " " in etctz:
                        etctz, dummy = etctz.split(" ", 1)
                    if "#" in etctz:
                        etctz, dummy = etctz.split("#", 1)
                    if not etctz:
                        continue

                    found_configs[tzpath] = etctz.replace(" ", "_")

        except (IOError, UnicodeDecodeError):
            # File doesn't exist or is a directory, or it's a binary file.
            continue

    # CentOS has a ZONE setting in /etc/sysconfig/clock,
    # OpenSUSE has a TIMEZONE setting in /etc/sysconfig/clock and
    # Gentoo has a TIMEZONE setting in /etc/conf.d/clock
    # We look through these files for a timezone:

    zone_re = re.compile(r"\s*ZONE\s*=\s*\"")
    timezone_re = re.compile(r"\s*TIMEZONE\s*=\s*\"")
    end_re = re.compile('"')

    for filename in ("etc/sysconfig/clock", "etc/conf.d/clock"):
        tzpath = os.path.join(_root, filename)
        try:
            with open(tzpath, "rt") as tzfile:
                data = tzfile.readlines()

            for line in data:
                # Look for the ZONE= setting.
                match = zone_re.match(line)
                if match is None:
                    # No ZONE= setting. Look for the TIMEZONE= setting.
                    match = timezone_re.match(line)
                if match is not None:
                    # Some setting existed
                    line = line[match.end():]
                    etctz = line[: end_re.search(line).start()]

                    # We found a timezone
                    found_configs[tzpath] = etctz.replace(" ", "_")

        except (IOError, UnicodeDecodeError):
            # UnicodeDecode handles when clock is symlink to /etc/localtime
            continue

    # systemd distributions use symlinks that include the zone name,
    # see manpage of localtime(5) and timedatectl(1)
    tzpath = os.path.join(_root, "etc/localtime")
    if os.path.exists(tzpath) and os.path.islink(tzpath):
        etctz = realtzpath = os.path.realpath(tzpath)
        start = etctz.find("/") + 1
        while start != 0:
            etctz = etctz[start:]
            try:
                pds.timezone(etctz)
                tzinfo = f"{tzpath} is a symlink to"
                found_configs[tzinfo] = etctz.replace(" ", "_")
            except pds.UnknownTimeZoneError:
                pass
            start = etctz.find("/") + 1

    if len(found_configs) > 0:
        # We found some explicit config of some sort!
        if len(found_configs) > 1:
            # Uh-oh, multiple configs. See if they match:
            unique_tzs = set()
            zoneinfo = os.path.join(_root, "usr", "share", "zoneinfo")
            directory_depth = len(zoneinfo.split(os.path.sep))

            for tzname in found_configs.values():
                # Look them up in /usr/share/zoneinfo, and find what they
                # really point to:
                path = os.path.realpath(os.path.join(zoneinfo, *tzname.split("/")))
                real_zone_name = "/".join(path.split(os.path.sep)[directory_depth:])
                unique_tzs.add(real_zone_name)

            if len(unique_tzs) != 1:
                message = "Multiple conflicting time zone configurations found:\n"
                for key, value in found_configs.items():
                    message += f"{key}: {value}\n"
                message += "Fix the configuration, or set the time zone in a TZ environment variable.\n"
                raise utils.ZoneInfoNotFoundError(message)

        # We found exactly one config! Use it.
        return list(found_configs.values())[0]


def _get_localzone(_root="/"):
    """Creates a timezone object from the timezone name.

    If there is no timezone config, it will try to create a file from the
    localtime timezone, and if there isn't one, it will default to UTC.

    The parameter _root makes the function look for files like /etc/localtime
    beneath the _root directory. This is primarily used by the tests.
    In normal usage you call the function without parameters."""

    # First try the ENV setting.
    tzenv = utils._tz_from_env()
    if tzenv:
        return tzenv

    tzname = _get_localzone_name(_root)
    if tzname is None:
        # No explicit setting existed. Use localtime
        for filename in ("etc/localtime", "usr/local/etc/localtime"):
            tzpath = os.path.join(_root, filename)

            if not os.path.exists(tzpath):
                continue
            with open(tzpath, "rb") as tzfile:
                tz = pds.wrap_zone(ZoneInfo.from_file(tzfile, key="local"))
                break
        else:
            warnings.warn("Can not find any timezone configuration, defaulting to UTC.")
            tz = timezone.utc
    else:
        tz = pds.timezone(tzname)

    if _root == "/":
        # We are using a file in etc to name the timezone.
        # Verify that the timezone specified there is actually used:
        utils.assert_tz_offset(tz)
    return tz


def get_localzone_name():
    """Get the computers configured local timezone name, if any."""
    global _cache_tz_name
    if _cache_tz_name is None:
        _cache_tz_name = _get_localzone_name()

    return _cache_tz_name


def get_localzone():
    """Get the computers configured local timezone, if any."""

    global _cache_tz
    if _cache_tz is None:
        _cache_tz = _get_localzone()

    return _cache_tz


def reload_localzone():
    """Reload the cached localzone. You need to call this if the timezone has changed."""
    global _cache_tz_name
    global _cache_tz
    _cache_tz_name = _get_localzone_name()
    _cache_tz = _get_localzone()

    return _cache_tz



from messagebus import (
    TokenPriority,
    TokenCreator,
)

nreports_group = "6F"


TK_NREPORTS_BUILD_REPORT = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=nreports_group,
    number="1",
)  # 116391937 // 0x6F000001



import logging
from typing import (
    Dict,
    List,
)

from model import Report
from model.exceptions import (
    ErrorBuildingReportData,
    ReportFileNotFound,
)
from nreports.build_report.dto import BuildReportRequest
from nreports.builders import DataBuilderFactory
from nreports.repository import I18N
from nreports_api.shared import JSON
from nreports_api.shared.report_dto import ReportDto

logger = logging.getLogger(name=__name__)


class BuildReportInteractor(object):

    def __init__(
        self,
        report_files: Dict[str, List[JSON]],
        data_builder_factory: DataBuilderFactory,
        i18n_repository: I18N,
    ) -> None:

        self.report_files = report_files
        self.data_builder_factory = data_builder_factory
        self.i18n_repository = i18n_repository

    def execute(
        self,
        request: BuildReportRequest,
    ) -> ReportDto:

        logger.info(msg=f"Report requested: {request.report_name}")

        report = self._get_report(report_name=request.report_name)

        data_to_build = report.get_data_to_build()

        data_builder = self.data_builder_factory.build(report_name=request.report_name)
        try:
            built_data = data_builder.build_data(
                data_to_build=data_to_build,
                **request.data,
            )
        except Exception as ex:
            raise ErrorBuildingReportData(
                report_name=request.report_name,
                error_message=str(ex),
            ) from ex

        formatted_report = report.format_data(
            built_data=built_data,
            i18n_repository=self.i18n_repository.build_i18n_repository(pos_id=request.pos_id or 0),
            l10n_repository=self.i18n_repository.build_l10n_repository(pos_id=request.pos_id or 0),
        )

        return ReportDto(formatted_report)

    def _get_report(
        self,
        report_name: str,
    ) -> Report:

        report = self.report_files.get(report_name.upper())
        if not report:
            raise ReportFileNotFound(report_name=report_name)

        return Report(report)



from dailygoals_services.model.repository import DailyGoalsRepository


class DeleteSalesChallengeInteractor:
    def __init__(self, daily_goals_repository):
        # type: (DailyGoalsRepository) -> None

        self.daily_goals_repository = daily_goals_repository

    def delete_sales_challenge(self, sale_challenge_id):
        # type: (int) -> int

        return self.daily_goals_repository.delete_sales_challenge(sale_challenge_id)



from typing import (
    List,
    Dict,
    Optional,
)

from devicemanager.model import (
    RequesterType,
    DeviceConfig,
)
from devicemanager.repository import DeviceManagerRepository


class DeviceStatusService(object):
    def __init__(
        self,
        device_configs,  # type: List[DeviceConfig]
    ):
        # type: (...) -> None

        self.device_configs = device_configs
        self.requester_current_device = None        # type: Optional[Dict[str, str]]
        self.status_dict = None                     # type: Optional[Dict[str, bool]]

    def get_enabled_device(
        self,               # type: DeviceStatusService
        requester_type,     # type: RequesterType
        requester_id,       # type: int
    ):
        key = self._get_requester_key(
            requester_type=requester_type.name,
            requester_id=requester_id,
        )
        return self.requester_current_device.get(key)

    def is_device_enabled(
        self,           # type: DeviceStatusService
        device_name,    # type: str
    ):
        # type: (...) -> bool

        if device_name in self.status_dict:
            return self.status_dict[device_name]
        return True

    def initialize_cache(
        self,        # type: DeviceStatusService
        repository,  # type: DeviceManagerRepository
    ):
        # type: (...) -> None

        self.status_dict = self._build_device_status_dict(
            repository=repository,
            device_configs=self.device_configs,
        )
        self.requester_current_device = self._build_requester_current_device(
            device_configs=self.device_configs,
            status_dict=self.status_dict,
        )

    def _build_requester_current_device(
        self,
        device_configs,  # type: List[DeviceConfig]
        status_dict,     # type: Dict[str, bool]
    ):
        # type: (...) -> Dict[str, str]

        requester_current_device = {}
        for config in device_configs:
            turned_on_device = self._get_turned_on_device(
                config=config,
                status_dict=status_dict,
            )
            key = self._get_requester_key(
                requester_type=config.requester_type,
                requester_id=config.requester_id,
            )
            requester_current_device[key] = turned_on_device

        return requester_current_device

    def _build_device_status_dict(
        self,            # type: DeviceStatusService
        repository,      # type: DeviceManagerRepository
        device_configs,  # type: List[DeviceConfig]
    ):
        # type: (...) -> Dict[str, bool]

        status_dict = self._build_status_dict_with_disabled_devices(repository=repository)
        self._add_all_devices_to_status_dict(
            device_configs=device_configs,
            status_dict=status_dict,
        )
        return status_dict

    @staticmethod
    def _build_status_dict_with_disabled_devices(
        repository,     # type: DeviceManagerRepository
    ):
        # type: (...) -> Dict[str, bool]

        ret = {}
        device_statuses = repository.get_disabled_devices()
        for device_name in device_statuses.keys():
            device = device_statuses[device_name]
            ret[device_name] = device.enabled
        return ret

    def _add_all_devices_to_status_dict(
        self,            # type: DeviceStatusService
        device_configs,  # type: List[DeviceConfig]
        status_dict,     # type: Dict[str, bool]
    ):
        # type: (...) -> None

        for device_config in device_configs:
            self._add_configured_devices_to_status_dict(
                config=device_config,
                status_dict=status_dict,
            )

    @staticmethod
    def _add_configured_devices_to_status_dict(
        config,         # type: DeviceConfig
        status_dict,    # type: Dict[str, bool]
    ):
        # type: (...) -> None

        device_order = [config.default_device] + config.fallback_devices
        for device_name in device_order:
            if device_name not in status_dict:
                status_dict[device_name] = True

    @staticmethod
    def _get_requester_key(
        requester_type,     # type: str
        requester_id,       # type: int
    ):
        # type: (...) -> str

        return str(requester_type) + str(requester_id)

    @staticmethod
    def _get_turned_on_device(
        config,       # type: DeviceConfig
        status_dict,  # type: Dict[str, bool]
    ):
        # type: (...) -> Optional[str]

        device_order = [config.default_device] + config.fallback_devices
        for device in device_order:
            if status_dict[device]:
                return device
        return None



from __future__ import print_function
from __future__ import absolute_import

import subprocess
import unittest

import greenlet
from . import _test_extension_cpp
from . import TestCase
from . import WIN

class CPPTests(TestCase):
    def test_exception_switch(self):
        greenlets = []
        for i in range(4):
            g = greenlet.greenlet(_test_extension_cpp.test_exception_switch)
            g.switch(i)
            greenlets.append(g)
        for i, g in enumerate(greenlets):
            self.assertEqual(g.switch(), i)

    def _do_test_unhandled_exception(self, target):
        import os
        import sys
        script = os.path.join(
            os.path.dirname(__file__),
            'fail_cpp_exception.py',
        )
        args = [sys.executable, script, target.__name__ if not isinstance(target, str) else target]
        __traceback_info__ = args
        with self.assertRaises(subprocess.CalledProcessError) as exc:
            subprocess.check_output(
                args,
                encoding='utf-8',
                stderr=subprocess.STDOUT
            )

        ex = exc.exception
        expected_exit = self.get_expected_returncodes_for_aborted_process()
        self.assertIn(ex.returncode, expected_exit)
        self.assertIn('fail_cpp_exception is running', ex.output)
        return ex.output


    def test_unhandled_nonstd_exception_aborts(self):
        # verify that plain unhandled throw aborts
        self._do_test_unhandled_exception(_test_extension_cpp.test_exception_throw_nonstd)

    def test_unhandled_std_exception_aborts(self):
        # verify that plain unhandled throw aborts
        self._do_test_unhandled_exception(_test_extension_cpp.test_exception_throw_std)

    @unittest.skipIf(WIN, "XXX: This does not crash on Windows")
    # Meaning the exception is getting lost somewhere...
    def test_unhandled_std_exception_as_greenlet_function_aborts(self):
        # verify that plain unhandled throw aborts
        output = self._do_test_unhandled_exception('run_as_greenlet_target')
        self.assertIn(
            # We really expect this to be prefixed with "greenlet: Unhandled C++ exception:"
            # as added by our handler for std::exception (see TUserGreenlet.cpp), but
            # that's not correct everywhere --- our handler never runs before std::terminate
            # gets called (for example, on arm32).
            'Thrown from an extension.',
            output
        )

    def test_unhandled_exception_in_greenlet_aborts(self):
        # verify that unhandled throw called in greenlet aborts too
        self._do_test_unhandled_exception('run_unhandled_exception_in_greenlet_aborts')


if __name__ == '__main__':
    unittest.main()



#
# The Python Imaging Library.
#
# QOI support for PIL
#
# See the README file for information on usage and redistribution.
#
from __future__ import annotations

import os

from . import Image, ImageFile
from ._binary import i32be as i32
from ._binary import o8


def _accept(prefix):
    return prefix[:4] == b"qoif"


class QoiImageFile(ImageFile.ImageFile):
    format = "QOI"
    format_description = "Quite OK Image"

    def _open(self):
        if not _accept(self.fp.read(4)):
            msg = "not a QOI file"
            raise SyntaxError(msg)

        self._size = tuple(i32(self.fp.read(4)) for i in range(2))

        channels = self.fp.read(1)[0]
        self._mode = "RGB" if channels == 3 else "RGBA"

        self.fp.seek(1, os.SEEK_CUR)  # colorspace
        self.tile = [("qoi", (0, 0) + self._size, self.fp.tell(), None)]


class QoiDecoder(ImageFile.PyDecoder):
    _pulls_fd = True

    def _add_to_previous_pixels(self, value):
        self._previous_pixel = value

        r, g, b, a = value
        hash_value = (r * 3 + g * 5 + b * 7 + a * 11) % 64
        self._previously_seen_pixels[hash_value] = value

    def decode(self, buffer):
        self._previously_seen_pixels = {}
        self._previous_pixel = None
        self._add_to_previous_pixels(b"".join(o8(i) for i in (0, 0, 0, 255)))

        data = bytearray()
        bands = Image.getmodebands(self.mode)
        while len(data) < self.state.xsize * self.state.ysize * bands:
            byte = self.fd.read(1)[0]
            if byte == 0b11111110:  # QOI_OP_RGB
                value = self.fd.read(3) + self._previous_pixel[3:]
            elif byte == 0b11111111:  # QOI_OP_RGBA
                value = self.fd.read(4)
            else:
                op = byte >> 6
                if op == 0:  # QOI_OP_INDEX
                    op_index = byte & 0b00111111
                    value = self._previously_seen_pixels.get(op_index, (0, 0, 0, 0))
                elif op == 1:  # QOI_OP_DIFF
                    value = (
                        (self._previous_pixel[0] + ((byte & 0b00110000) >> 4) - 2)
                        % 256,
                        (self._previous_pixel[1] + ((byte & 0b00001100) >> 2) - 2)
                        % 256,
                        (self._previous_pixel[2] + (byte & 0b00000011) - 2) % 256,
                    )
                    value += (self._previous_pixel[3],)
                elif op == 2:  # QOI_OP_LUMA
                    second_byte = self.fd.read(1)[0]
                    diff_green = (byte & 0b00111111) - 32
                    diff_red = ((second_byte & 0b11110000) >> 4) - 8
                    diff_blue = (second_byte & 0b00001111) - 8

                    value = tuple(
                        (self._previous_pixel[i] + diff_green + diff) % 256
                        for i, diff in enumerate((diff_red, 0, diff_blue))
                    )
                    value += (self._previous_pixel[3],)
                elif op == 3:  # QOI_OP_RUN
                    run_length = (byte & 0b00111111) + 1
                    value = self._previous_pixel
                    if bands == 3:
                        value = value[:3]
                    data += value * run_length
                    continue
                value = b"".join(o8(i) for i in value)
            self._add_to_previous_pixels(value)

            if bands == 3:
                value = value[:3]
            data += value
        self.set_as_raw(bytes(data))
        return -1, 0


Image.register_open(QoiImageFile.format, QoiImageFile, _accept)
Image.register_decoder("qoi", QoiDecoder)
Image.register_extension(QoiImageFile.format, ".qoi")
